"""
General structure of a problem formulation
"""
from abc import ABC
import math
from enum import Enum
from typing import Optional, Union, Iterable, Tuple, Dict, Callable, Collection, List
from unocg.materials import Material
import numpy as np
import scipy as sp
import torch
from scipy import sparse
from tqdm import tqdm


class BC(Enum):
    PERIODIC = 0
    DIRICHLET_TB = 1
    DIRICHLET_LR = 2
    DIRICHLET = 3
    DIRICHLET_TB_X = 4
    DIRICHLET_TB_Y = 5
    DIRICHLET_LR_X = 6
    DIRICHLET_LR_Y = 7
    DIRICHLET_X = 8
    DIRICHLET_Y = 9
    DIRICHLET_FB = 10
    NEUMANN = 11

class FieldType(Enum):
    NODE = 0
    CELL = 1
    HOM = 2
    GAUSS = 3


class Problem:
    def __init__(self, n_grid: Union[torch.Size, Collection[int]], bc: Optional[BC] = None, material: Optional[Material] = None, quad_degree: int = 2, lengths = None):
        if bc is None:
            self.bc = BC.PERIODIC
        else:
            self.bc = bc
        self._n_dim = len(n_grid)
        self._n_channels = 1
        self.dtype = torch.float64
        self._n_grid = n_grid
        self.dof_shape = self._n_grid
        self._n_dof = None
        self._loadings = None
        self.quad_degree = quad_degree
        self.material = material
        if lengths is None:
            self.lengths = torch.ones(self.n_dim, dtype=self.dtype)
        else:
            self.lengths = torch.tensor(lengths, dtype=self.dtype)
        self.h = self.lengths / torch.tensor(self.n_grid, dtype=self.dtype)

    @property
    def n_dim(self) -> int:
        return self._n_dim

    @property
    def dims(self):
        return self.n_dim * (slice(None),)

    @property
    def dims_list(self):
        return tuple(range(-1, -(1 + self.n_dim), -1))

    @property
    def expand_dims(self):
        return self.n_dim * (None,)

    @property
    def einsum_dims(self):
        if self.n_dim == 2:
            return "yx"
        elif self.n_dim == 3:
            return "zyx"
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")

    @property
    def n_channels(self) -> int:
        return self._n_channels

    @property
    def ch_dim(self):
        """
        Channel dimension

        :return:
        """
        return -(1 + self.n_dim)

    @property
    def n_grid(self) -> Union[torch.Size, Collection[int]]:
        return self._n_grid

    @property
    def n_dof(self) -> int:
        return self._n_dof

    @property
    def loadings(self) -> torch.Tensor:
        return self._loadings

    @loadings.setter
    def loadings(self, loadings: torch.Tensor):
        self._loadings = loadings

    @property
    def n_loadings(self):
        return self.loadings.shape[0]

    def get_vec_shape(self, param_fields: torch.Tensor, loadings: Optional[torch.Tensor] = None) -> torch.Size:
        """
        Get shape for DOF vectors

        :param param_fields:
        :param loadings:
        :return:
        """
        if loadings is None:
            n_loadings = self.n_loadings
        else:
            n_loadings = loadings.shape[0]

        batch_shape = param_fields.shape[: -(self.n_dim + 1)]
        return torch.Size((*batch_shape, n_loadings, self.n_dof))

    def get_field_shape(self, param_fields, loadings: Optional[torch.Tensor] = None):
        """
        Get shape for fields

        :param param_fields:
        :param loadings:
        :return:
        """
        if loadings is None:
            n_loadings = self.n_loadings
        else:
            n_loadings = loadings.shape[0]

        batch_shape = param_fields.shape[:self.ch_dim]

        return torch.Size((*batch_shape, n_loadings, self.n_channels, *self.dof_shape))

    def reshape_vec(self, batch: Union[torch.Tensor, np.ndarray]) -> Union[torch.Tensor, np.ndarray]:
        """
        Reshape batch of fields to batch of DOF vectors

        :param batch:
        :return:
        """
        batch_shape = batch.shape[: -(self.n_dim + 1)]
        return batch.reshape((*batch_shape, self.n_dof))

    def reshape_field(self, batch: Union[torch.Tensor, np.ndarray]) -> Union[torch.Tensor, np.ndarray]:
        """
        Reshape batch of DOF vectors to batch of fields

        :param batch:
        :return:
        """
        batch_shape = batch.shape[:-1]
        return batch.reshape((*batch_shape, self.n_channels, *self.dof_shape))

    def compute_residual(self,
                         u: torch.Tensor,
                         param_fields: Optional[torch.Tensor] = None,
                         loadings: Optional[torch.Tensor] = None,
                         state_fields: Optional[torch.Tensor] = None) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Compute residual

        :param u:
        :param param_fields:
        :param loadings:
        :param state_fields:
        :return:
        """
        raise NotImplementedError()

    def matvec(self,
               d: torch.Tensor,
               param_fields: Optional[torch.Tensor] = None,
               state_fields: Optional[torch.Tensor] = None) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Matrix-vector product with stiffness matrix

        :param d:
        :param param_fields:
        :param state_fields:
        :return:
        """
        raise NotImplementedError()

    def compute_field(
            self,
            u: torch.Tensor,
            param_fields: Optional[torch.Tensor] = None,
            loadings: Optional[torch.Tensor] = None,
            state_fields: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Compute field based on DOF vector u

        :param u:
        :param param_fields:
        :param loadings:
        :param state_fields:
        :return:
        """
        raise NotImplementedError()
    
    def get_hom_response(
            self,
            u: torch.Tensor,
            param_fields: Optional[torch.Tensor] = None,
            loadings: Optional[torch.Tensor] = None,
            state_fields: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Compute homogenized response based on DOF vector u

        :param u:
        :param param_fields:
        :param loadings:
        :param state_fields:
        :return:
        """
        raise NotImplementedError()

    def split_fields(self, fields: torch.Tensor) -> Dict[str, Tuple[FieldType, torch.Tensor]]:
        """
        Split general field into problem-specific physically meaningful fields

        :param fields:
        :return:
        """
        raise NotImplementedError()

    def get_node_coords(self):
        """

        :return:
        """
        if self.n_dim == 2:
            ny, nx = self.n_grid
            ly, lx = self.lengths
            x = torch.linspace(0.0, lx, nx + 1)
            y = torch.linspace(0.0, ly, ny + 1)
            Y, X = torch.meshgrid(y, x, indexing="ij")
            vertices = torch.stack([X, Y], dim=-3)
        elif self.n_dim == 3:
            nz, ny, nx = self.n_grid
            lz, ly, lx = self.lengths
            x = torch.linspace(0.0, lx, nx + 1)
            y = torch.linspace(0.0, ly, ny + 1)
            z = torch.linspace(0.0, lz, nz + 1)
            Z, Y, X = torch.meshgrid(z, y, x, indexing="ij")
            vertices = torch.stack([X, Y, Z], dim=-4)
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")
        return vertices

    def get_dof_coords(self):
        """

        :return:
        """
        if self.n_dim == 2:
            ny, nx = self.n_grid
            vertices = self.get_node_coords()
            if self.bc == BC.PERIODIC:
                dof_vertices = vertices[...,1:,1:]
            elif self.bc == BC.DIRICHLET:
                dof_vertices = vertices[...,1:-1,1:-1]
            else:
                raise NotImplementedError(f"BC {self.bc} is not supported")
        elif self.n_dim == 3:
            nz, ny, nx = self.n_grid
            vertices = self.get_node_coords()
            if self.bc == BC.PERIODIC:
                dof_vertices = vertices[...,1:,1:,1:]
            elif self.bc == BC.DIRICHLET:
                dof_vertices = vertices[...,1:-1,1:-1,1:-1]
            else:
                raise NotImplementedError(f"BC {self.bc} is not supported")
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")
        return dof_vertices


    def get_cell_coords(self):
        """

        :return:
        """
        if self.n_dim == 2:
            ny, nx = self.n_grid
            ly, lx = self.lengths
            x_c = torch.linspace(0.5 / nx, lx - 0.5 / nx, nx)
            y_c = torch.linspace(0.5 / ny, ly - 0.5 / ny, ny)
            Y_c, X_c = torch.meshgrid(y_c, x_c, indexing="ij")
            vertices_c = torch.stack([X_c, Y_c], dim=-3).flatten(start_dim=1).T
        elif self.n_dim == 3:
            nz, ny, nx = self.n_grid
            lz, ly, lx = self.lengths
            x_c = torch.linspace(0.5 / nx, lx - 0.5 / nx, nx)
            y_c = torch.linspace(0.5 / ny, ly - 0.5 / ny, ny)
            z_c = torch.linspace(0.5 / nz, lz - 0.5 / nz, nz)
            Z_c, Y_c, X_c = torch.meshgrid(z_c, y_c, x_c, indexing="ij")
            vertices_c = torch.stack([X_c, Y_c, Z_c], dim=-4).flatten(start_dim=1).T
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")
        return vertices_c

    @staticmethod
    def get_scatter_matrix_sp(dof: torch.Tensor, shape=None):
        """

        :param dof:
        :param shape:
        :return:
        """
        if shape is None:
            shape = (dof.shape[0], dof.shape[0])

        co = np.stack([np.arange(dof.shape[0]), dof])
        da = np.ones(dof.shape[0])
        L = sp.sparse.coo_array((da, co), shape=shape).tocsr()
        return L

    @staticmethod
    def get_scatter_matrix(dof: torch.Tensor, shape=None):
        """

        :param dof:
        :param shape:
        :return:
        """
        if shape is None:
            shape = (dof.shape[0], dof.shape[0])

        co = torch.stack([torch.arange(dof.shape[0], dtype=int, device=dof.device), dof])
        da = torch.ones(dof.shape[0], device=dof.device)
        L = torch.sparse_coo_tensor(co, da, size=shape).coalesce()
        return L

    def get_quad_rule(self):
        if self.quad_degree == 1:
            rule1d = torch.tensor([0.5], dtype=self.dtype)
            rule1dw = torch.tensor([1.0], dtype=self.dtype)
        elif self.quad_degree == 2:
            rule1d = torch.tensor([0.5 - math.sqrt(3.0) / 6.0, 0.5 + math.sqrt(3.0) / 6.0], dtype=self.dtype)
            rule1dw = torch.tensor([0.5, 0.5], dtype=self.dtype)
        elif self.quad_degree == 3:
            rule1d = torch.tensor([0.5 - math.sqrt(0.6) / 2.0, 0.5 + math.sqrt(0.6) / 2.0], dtype=self.dtype)
            rule1dw = torch.tensor([5. / 18., 4. / 9., 5. / 18.], dtype=self.dtype)
        else:
            raise NotImplementedError(f"Quadrature degree {self.quad_degree} is not supported")
        return rule1d, rule1dw

    @staticmethod
    def extract_image(input_field: torch.Tensor) -> torch.Tensor:
        """
        Extract image

        :param input_field:
        :return:
        """
        return input_field[..., 0, :, :]

    @staticmethod
    def extract_param_fields(input_field: torch.Tensor) -> torch.Tensor:
        """
        Extract parameter fields

        :param input_field:
        :return:
        """
        return input_field[..., 1:, :, :]

    @staticmethod
    def extract_parameters(input_field: torch.Tensor) -> torch.Tensor:
        """
        Extract parameters

        :param input_field:
        :return: params
        """
        if input_field.ndim == 3:
            input_field = torch.unsqueeze(input_field, dim=0)

        params_view = input_field.view(*input_field.shape[:-2], -1)
        idx0 = params_view[..., 0, :].argmin(dim=-1)
        idx1 = params_view[..., 0, :].argmax(dim=-1)
        params0 = params_view[torch.arange(input_field.shape[0]), 1:, idx0]
        params1 = params_view[torch.arange(input_field.shape[0]), 1:, idx1]
        params = torch.stack([params0, params1], dim=-2)
        return params

    def extract_dof(self, field: torch.Tensor) -> torch.Tensor:
        """
        Extract DOF vector from field solution

        :param field:
        :return:
        """
        if self.bc == BC.PERIODIC:
            dof = field[..., :-1, :-1].ravel()
        elif self.bc == BC.DIRICHLET_LR:
            dof = field[..., :-1, 1:-1].ravel()
        elif self.bc == BC.DIRICHLET_TB:
            dof = field[..., 1:-1, :-1].ravel()
        elif self.bc == BC.DIRICHLET:
            dof = field[..., 1:-1, 1:-1].ravel()
        else:
            raise NotImplementedError()
        return dof

    def get_param_fields(self, image, params):
        """
        Construct parameter fields for a two-phase microstructure described by image based on params.

        :param image:
        :param params:
        :return:
        """
        param_fields = torch.einsum("..." + self.einsum_dims + ",i->...i" + self.einsum_dims, (image.to(dtype=int) == 0).to(dtype=params.dtype), params[0])
        for i in range(1, params.shape[0]):
            param_fields += torch.einsum("..." + self.einsum_dims + ",i->...i" + self.einsum_dims, (image.to(dtype=int) == i).to(dtype=params.dtype), params[i])
        return param_fields

    def homogenize(self, field: torch.Tensor) -> torch.Tensor:
        """
        Homogenization of a tensor field on a 2d grid by volume averaging

        :param field: tensor field with shape [..., n, n]
        :return:
        """
        return field.nanmean(self.dims_list)

    def field_norm(self, field: torch.Tensor) -> torch.Tensor:
        """
        Compute the point-wise norm of a tensor field

        :param field:
        :return: point-wise norm
        """
        return torch.sqrt(torch.square(field).nansum(self.ch_dim))

    def hom_norm(self, field: torch.Tensor) -> torch.Tensor:
        """
        Compute the homogenized norm of a tensor field

        :param field:
        :return: homogenized norm
        """
        return self.field_norm(field).nanmean(self.dims_list)

    def zero_mean(self, u: torch.Tensor) -> torch.Tensor:
        """
        Enforce zero-mean property for solution vector u
        """
        field = self.reshape_field(u)
        field_zero_mean = field - field.mean(self.dims_list)[..., *self.expand_dims]
        return self.reshape_vec(field_zero_mean)

    @classmethod
    def pad_circular_nd(cls, x: torch.Tensor, pad: int, dim) -> torch.Tensor:
        """
        Periodic padding for tensor fields on a 2d grid

        :param x: shape [H, W] of the grid
        :param pad: int >= 0
        :param dim: the dimension over which the tensors are padded
        :return:
        """
        if isinstance(dim, int):
            dim = [dim]

        for d in dim:
            if d >= len(x.shape):
                raise IndexError(f"dim {d} out of range")

            idx = tuple(slice(0, None if s != d else pad, 1) for s in range(len(x.shape)))
            x = torch.cat([x, x[idx]], dim=d)

            idx = tuple(slice(None if s != d else -2 * pad, None if s != d else -pad, 1) for s in range(len(x.shape)))
            x = torch.cat([x[idx], x], dim=d)
            pass

        return x

    @classmethod
    def pad_circular_2d(cls, x: torch.Tensor, pad_left: int, pad_right: int, mode: str = "circular") -> torch.Tensor:
        """
        Periodic padding for tensor fields on a 2d grid

        :param x: shape [H, W] of the grid
        :param pad_left: int >= 0
        :param pad_right: int >= 0
        :param mode: str {'circular', 'constant', 'reflect', 'replicate'}
        :return:
        """
        paddings = [pad_left, pad_right, pad_left, pad_right]
        original_shape = torch.tensor(x.shape)
        padded_shape = original_shape.clone()
        padded_shape[-2:] += torch.tensor([pad_left + pad_right, pad_left + pad_right])
        return torch.nn.functional.pad(x.reshape(-1, *original_shape[-3:]), paddings, mode=mode).reshape(*padded_shape)

    def get_image_masks(self, image: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Get index masks for

        - interior of matrix (interior_matr),
        - interface of matrix (interface_matr),
        - interface of inclusion (interface_incl),
        - interior of inclusion (interior_incl).

        :param image:
        :type image: torch.Tensor
        """
        # TODO: generalize for 3D
        image = image.unsqueeze(dim=self.ch_dim)
        kernel_tensor = torch.ones((1, 1, 3, 3)).to(device=image.device, dtype=image.dtype)
        img_erosion = 1 - torch.clamp(torch.nn.functional.conv2d(1 - image, kernel_tensor, padding=(1, 1)), 0, 1)
        img_dilation = torch.clamp(torch.nn.functional.conv2d(image, kernel_tensor, padding=(1, 1)), 0, 1)
        interface_incl = (image - img_erosion).bool().squeeze(dim=self.ch_dim)
        interface_matr = (img_dilation - image).bool().squeeze(dim=self.ch_dim)
        interior_incl = torch.clamp(img_erosion - interface_incl.int(), 0, 1).bool().squeeze(dim=self.ch_dim)
        interior_matr = (1 - img_dilation).bool().squeeze(dim=self.ch_dim)
        return interior_matr, interface_matr, interface_incl, interior_incl

    @staticmethod
    def get_complex_dtype(dtype: torch.dtype) -> torch.dtype:
        """

        :param dtype:
        :return:
        """
        complex_dtypes = {
            torch.float16: torch.complex32,
            torch.half: torch.complex32,
            torch.complex32: torch.complex32,
            torch.float32: torch.complex64,
            torch.float: torch.complex64,
            torch.complex64: torch.complex64,
            torch.cfloat: torch.complex64,
            torch.float64: torch.complex128,
            torch.double: torch.complex128,
            torch.complex128: torch.complex128,
            torch.cdouble: torch.complex128,
        }

        try:
            for dtype_normal, dtype_complex in complex_dtypes.items():
                if dtype_normal == dtype:
                    return dtype_complex
        except:
            raise ValueError("Unknown type")
        raise ValueError("Unknown type")

    @staticmethod
    def get_complex_dtype_sp(dtype: torch.dtype) -> torch.dtype:
        """

        :param dtype:
        :return:
        """
        complex_dtypes = {
            np.float32: np.complex64,
            np.complex64: np.complex64,
            np.cfloat: np.complex128,
            "float64": np.complex128,
            np.float64: np.complex128,
            np.double: np.complex128,
            np.complex128: np.complex128,
            np.cdouble: np.complex128,
        }

        try:
            for dtype_normal, dtype_complex in complex_dtypes.items():
                if dtype_normal == dtype:
                    return dtype_complex
        except:
            raise ValueError("Unknown type")
        raise ValueError("Unknown type")

    def get_shapes(
            self, input_field: torch.Tensor, channels: Optional[int] = 1, n_loadings: Optional[int] = 2
    ) -> Tuple[Iterable, Iterable, Iterable]:
        """

        :param input_field:
        :param channels:
        :param n_loadings:
        :return:
        """
        batch_shape = input_field.shape[:self.ch_dim]
        field_shape = self.get_field_shape(input_field)
        vec_shape = self.get_vec_shape(input_field)
        return batch_shape, field_shape, vec_shape

    def extract_shape(self, input_field: torch.Tensor) -> torch.Size:
        """
        Extract shape of discretization from input

        :param input_field:
        :return: shape of discretization
        """
        return input_field.shape[-self.n_dim:]

    @classmethod
    def fft2_real(cls, field: torch.Tensor) -> torch.Tensor:
        """
        Computes FFT of real-valued field using rfft2(field) and reconstructs the negative frequencies
        to match the output of fft2(field)

        :param field:
        :return:
        """
        # TODO: generalize or remove
        assert field.shape[-1] == field.shape[-2]
        n = field.shape[-1]
        nh = n // 2

        rfft_field = torch.fft.rfft2(field)
        fft_field = torch.zeros_like(field, dtype=cls.get_complex_dtype(field.dtype))
        fft_field[..., :, : (nh + 1)] = rfft_field  # 0th and positive frequencies from real-valued fft

        # reconstruct negative frequencies
        fft_field[..., -nh:, -nh:] = torch.flip(rfft_field[..., 1: (nh + 1), 1: (nh + 1)].conj(), dims=[-1, -2])
        fft_field[..., 1: (nh + 1), -nh:] = torch.flip(rfft_field[..., -nh:, 1: (nh + 1)].conj(), dims=[-1, -2])
        fft_field[..., 0, -nh:] = torch.flip(rfft_field[..., 0, 1: (nh + 1)].conj(), dims=[-1])
        return fft_field

    @classmethod
    def dft_matrix(cls, shape):
        # TODO: generalize or remove
        return torch.fft.fft(torch.eye(shape[-1]))

    def get_coordinates(self, image: torch.Tensor) -> torch.Tensor:
        """
        Get coordinates of all nodes in the image

        :param image: tensor with shape (..., H, W)
        :param lx: length of x direction
        :param ly: length of y direction
        :return: coordinates with shape (C, H, W); C=2
        """
        #return self.get_node_coords().to(dtype=image.dtype, device=image.device)
        if self.n_dim == 2:
            ny, nx = image.shape[-2:]
            ly, lx = self.lengths
            x = torch.linspace(-lx / 2.0, lx / 2.0 - 1.0 / nx, nx, dtype=image.dtype, device=image.device)
            y = torch.linspace(-ly / 2.0, ly / 2.0 - 1.0 / ny, ny, dtype=image.dtype, device=image.device)
            Y, X = torch.meshgrid(y, x, indexing="ij")
            return torch.stack([X, Y], dim=self.ch_dim)
        elif self.n_dim == 3:
            nz, ny, nx = image.shape[-3:]
            lz, ly, lx = self.lengths
            x = torch.linspace(-lx / 2.0, lx / 2.0 - 1.0 / nx, nx, dtype=image.dtype, device=image.device)
            y = torch.linspace(-ly / 2.0, ly / 2.0 - 1.0 / ny, ny, dtype=image.dtype, device=image.device)
            z = torch.linspace(-lz / 2.0, lz / 2.0 - 1.0 / nz, nz, dtype=image.dtype, device=image.device)
            Z, Y, X = torch.meshgrid(z, y, x, indexing="ij")
            return torch.stack([X, Y, Z], dim=self.ch_dim)

    def mandel_to_matrix(self, tensor_mandel: torch.Tensor) -> torch.Tensor:
        """

        :param tensor_mandel: strain/stress tensor in mandel notation with shape (..., 3)
        :return: tensor in matrix notation with shape (..., 2, 2)
        """
        if self.n_dim == 2:
            tensor_matrix = torch.stack(
                [
                    torch.stack([tensor_mandel[..., 0], tensor_mandel[..., 2] / math.sqrt(2.0)], dim=-1),
                    torch.stack([tensor_mandel[..., 2] / math.sqrt(2.0), tensor_mandel[..., 1]], dim=-1),
                ],
                dim=-1,
            )
        elif self.n_dim == 3:
            tensor_matrix = torch.stack(
                [
                    torch.stack([tensor_mandel[..., 0], tensor_mandel[..., 5] / math.sqrt(2.0), tensor_mandel[..., 4] / math.sqrt(2.0)], dim=-1),
                    torch.stack([tensor_mandel[..., 5] / math.sqrt(2.0), tensor_mandel[..., 1], tensor_mandel[..., 3] / math.sqrt(2.0)], dim=-1),
                    torch.stack([tensor_mandel[..., 4] / math.sqrt(2.0), tensor_mandel[..., 3] / math.sqrt(2.0), tensor_mandel[..., 2]], dim=-1),
                ],
                dim=-1,
            )
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported for this operation")
        return tensor_matrix

    def get_boundary_idx(self, shape: Union[torch.Size, Iterable[int]]) -> Tuple[torch.Tensor, torch.Tensor]:
        """

        :param shape:
        :return:
        """
        # TODO: generalize or remove
        x = torch.arange(shape[-1])
        y = torch.arange(shape[-2])
        Y, X = torch.meshgrid(y, x, indexing="ij")
        idx = torch.stack([X, Y], dim=0)
        boundary_idx = torch.cat(
            [idx[:, 0, :], idx[:, :, -1][..., 1:], idx[:, -1, :][..., :-1].flip(-1), idx[:, :, 0].flip(-1)], dim=-1
        )
        return boundary_idx[0].ravel(), boundary_idx[1].ravel()

    def generate_circle_ms(
            self,
            r: float = 0.4,
            cx: float = 0.0,
            cy: float = 0.0,
            cz: float = 0.0,
    ) -> torch.Tensor:
        """
        Generate microstructure of given shape with spherical inclusion with center (cx, cy) and with radius r

        :param shape:
        :param r:
        :param cx:
        :param cy:
        :param lx:
        :param ly:
        :return: microstructure
        """
        # TODO: generalize or remove
        if self.n_dim == 2:
            ny, nx = self.n_grid[-2:]
            ly, lx = self.lengths[-2:]
            hx, hy = lx / nx, ly / ny
            x = torch.linspace(-lx / 2.0 + hx / 2.0, lx / 2.0 - hx / 2.0, nx)
            y = torch.linspace(-ly / 2.0 + hy / 2.0, ly / 2.0 - hy / 2.0, ny)
            Y, X = torch.meshgrid(y, x, indexing="ij")
            ms = ((X - cx) ** 2 + (Y - cy) ** 2 <= r ** 2).to(dtype=torch.int)
            return ms
        elif self.n_dim == 3:
            nz, ny, nx = self.n_grid[-3:]
            lz, ly, lx = self.lengths[-3:]
            hx, hy, hz = lx / nx, ly / ny, lz / nz
            x = torch.linspace(-lx / 2.0 + hx / 2.0, lx / 2.0 - hx / 2.0, nx)
            y = torch.linspace(-ly / 2.0 + hy / 2.0, ly / 2.0 - hy / 2.0, ny)
            z = torch.linspace(-lz / 2.0 + hz / 2.0, lz / 2.0 - hz / 2.0, nz)
            Z, Y, X = torch.meshgrid(z, y, x, indexing="ij")
            ms = ((X - cx) ** 2 + (Y - cy) ** 2 + (Z - cz) ** 2 <= r ** 2).to(dtype=torch.int)
            return ms
        else:
            raise NotImplementedError(f"Microstructure generation for dimension {self.n_dim} is not supported")

    def generate_ellipsoid_ms(self, ellipses):
        """
        Create an (nx, ny) image with ellipses.

        Args:
            nx (int): image height (rows)
            ny (int): image width (columns)
            ellipses (list of dict): list of ellipses,
                each dict must contain:
                - "center": (x0, y0) ellipse center
                - "axes": (a, b) semi-major and semi-minor axes
                - "angle": theta (rotation angle in radians)
                - "value": fill value (default=1.0)
            device: torch device ("cpu" or "cuda")
            dtype: tensor dtype

        Returns:
            torch.Tensor: (nx, ny) image with ellipses
        """
        if self.n_dim != 2:
            raise NotImplementedError()

        nx, ny = self.n_grid

        # Coordinate grid
        ys = torch.arange(nx).view(-1, 1).expand(nx, ny)
        xs = torch.arange(ny).view(1, -1).expand(nx, ny)
        
        # Empty image
        img = torch.zeros((nx, ny))
        
        for e in ellipses:
            x0, y0 = e["center"]
            a, b = e["axes"]
            theta = torch.tensor(e.get("angle", 0.0))
            val = e.get("value", 1.0)

            # Periodic coordinate differences
            dx = ((xs - x0 + ny/2) % ny) - ny/2
            dy = ((ys - y0 + nx/2) % nx) - nx/2

            # Rotate coordinates
            cos_t = torch.cos(theta)
            sin_t = torch.sin(theta)
            x_rot = dx * cos_t + dy * sin_t
            y_rot = -dx * sin_t + dy * cos_t

            # Ellipse mask
            mask = (x_rot**2) / (a**2) + (y_rot**2) / (b**2) <= 1.0

            # Overwrite pixels
            img = torch.where(mask, val * torch.ones_like(img), img)
    
        return img

    @staticmethod
    def torch_sparse_to_sp(tensor: torch.Tensor) -> sparse.coo_array:
        """
        Convert PyTorch sparse tensor to scipy sparse array

        :param tensor:
        :return:
        """
        tensor = tensor.to_sparse_coo().coalesce()
        sp_array = sparse.coo_array(
            (tensor.values().detach().cpu().numpy(), (tensor.indices()[0].detach().cpu().numpy(), tensor.indices()[1].detach().cpu().numpy())),
            shape=tensor.shape
        )
        return sp_array

    """
    def petsc_sparse_to_sp(mat: PETSc.Mat) -> sparse.coo_array:

        Convert PETSc sparse tensor to scipy sparse array

        :param tensor:
        :return:

        tensor = tensor.to_sparse_coo().coalesce()
        sp_array = sparse.coo_array(
            (tensor.values().numpy(), (tensor.indices()[0].numpy(), tensor.indices()[1].numpy())), shape=tensor.shape
        )
        return sp_array
    """

    @staticmethod
    def sp_sparse_to_torch(sp_array: sparse.coo_array) -> torch.Tensor:
        """
        Convert scipy sparse array to PyTorch sparse tensor

        :param sp_array:
        :return:
        """
        sp_array = sp_array.tocoo()
        values = sp_array.data
        indices = np.vstack((sp_array.row, sp_array.col))

        i = torch.LongTensor(indices)
        v = torch.tensor(values)
        shape = sp_array.shape

        sparse_torch = torch.sparse_coo_tensor(i, v, torch.Size(shape))
        return sparse_torch

    @staticmethod
    def sparse_diag(sparse_tensor: torch.Tensor, return_sparse: bool = False) -> torch.Tensor:
        """
        Extract diagonal of sparse tensor as PyTorch has no native implementation for it so far

        :param sparse_tensor: PyTorch sparse tensor
        :param return_sparse: If a sparse diagonal matrix should be returned of a vector containing only the diagonal
        :return:
        """
        assert sparse_tensor.ndim == 2
        assert sparse_tensor.shape[-1] == sparse_tensor.shape[-2]

        sparse_tensor = sparse_tensor.to_sparse_coo().coalesce()
        idx = sparse_tensor.indices()
        val = sparse_tensor.values()
        diag_mask = idx[0, :] == idx[1, :]
        idx_diag = idx[:, diag_mask]
        val_diag = val[diag_mask]

        if return_sparse:
            diag = torch.sparse_coo_tensor(idx_diag, val_diag)
        else:
            diag = torch.zeros(sparse_tensor.shape[-1], device=sparse_tensor.device)
            diag[idx_diag[0, :]] = val_diag

        return diag

    @staticmethod
    def matvec_to_sparse(
            shape: Union[Tuple[int, int], torch.Size],
            matvec: Callable[[torch.Tensor], torch.Tensor],
            eps: float = 1e-16,
            progress: bool = False,
    ) -> torch.Tensor:
        """
        Converts a matrix-free matvec function to the corresponding sparse PyTorch matrix

        :param shape: shape of the matrix
        :param matvec: matrix-free matvec function
        :param eps: tolerance for non-zero entries
        :param progress: show progress bar
        :return:
        """
        assert shape[-2] == shape[-1]
        N = shape[-1]
        idx = torch.arange(N, dtype=torch.int)
        indices = torch.zeros((2, 0), dtype=torch.int)
        values = torch.tensor([])

        for i in tqdm(idx, disable=not progress):
            u = torch.zeros(N)
            u[i] = 1.0
            r = matvec(u)
            mask = torch.abs(r) > 1e-16
            idx_x = torch.repeat_interleave(torch.unsqueeze(i, 0), mask.sum())
            idx_y = idx[mask.ravel()]
            current_indices = torch.stack([idx_x, idx_y])
            indices = torch.cat([indices, current_indices], dim=-1)
            values = torch.cat([values, r[mask].ravel()], dim=-1)
        A_sparse = torch.sparse_coo_tensor(indices, values).coalesce()
        return A_sparse

    @staticmethod
    def setdiff(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        """

        :param a:
        :param b:
        :return:
        """
        combined = torch.cat((a.unique(), b.unique()))
        uniques, counts = combined.unique(return_counts=True)
        difference = uniques[counts == 1]
        intersection = uniques[counts > 1]
        return difference
    
    def get_deformations(self, disp, loading, fluctuation_scaling: float = 1.0):
        coords = self.get_coordinates(disp)
        loading_matrix = self.mandel_to_matrix(loading).to(dtype=disp.dtype, device=disp.device)
        deformations = torch.einsum("...ij,j" + self.einsum_dims + "->...i" + self.einsum_dims, loading_matrix, coords) + fluctuation_scaling * disp
        return deformations

    def get_deformed_coordinates(
        self, disp: torch.Tensor, loading, lx=1.0, ly=1.0, fluctuation_scaling: float = 1.0, deformation_scaling: float = 1.0
    ) -> torch.Tensor:
        """
        Get coordinates in deformed configuration according to displacement fluctuations disp and loading
        :param disp: tensor with shape of (..., L, 2, H, W)
        :param loading: strain loading in mandel notation with shape (L, 3)
        :param lx: length along x-axis
        :param ly: length along y-axis
        :param fluctuation_scaling: scaling factor of displacement fluctuations, defaults to 1.0
        :param deformation_scaling: scaling factor of displacement fluctuations, defaults to 1.0
        :return: coordinates with shape (..., L, C, H, W); C=2
        """
        coords = self.get_coordinates(disp)
        loading_matrix = self.mandel_to_matrix(loading).to(dtype=disp.dtype, device=disp.device)
        deformations = torch.einsum("...ij,jyx->...iyx", loading_matrix, coords) + fluctuation_scaling * disp
        deformed_coords = coords + deformation_scaling * deformations
        return deformed_coords



"""
Plotting utilities
"""
from typing import Optional, Iterable, List, Tuple

import matplotlib
from matplotlib.tri import Triangulation
import numpy as np
import torch
from matplotlib import pyplot as plt, ticker
import math
import shutil
from unocg.problems import Problem, BC

try:
    import pyvista as pv
except:
    pass

plt.rcParams["text.usetex"] = True if shutil.which('latex') else False
matplotlib.rcParams["text.latex.preamble"] = r"\usepackage{amsmath}"
plot_width = 6.3

colors = [
    "g",
    "#00BEFF",
    "#004191",
    "r",
    "k"
]


def plot_ms(image: torch.Tensor, file: Optional[str] = None, show_axis: bool = False, show_cbar: bool = False):
    """
    Plot microstructure image

    :param image:
    :param file:
    :param show_axis:
    :param show_cbar:
    :return:
    """
    image = image.detach().cpu()
    fig, ax = plt.subplots(1, 1, figsize=[6.3 / 2, 2.65 if show_cbar else 6.3 / 2], dpi=300)
    ms_cmap = plt.get_cmap("viridis", 2)
    im = ax.imshow(image, origin="lower", interpolation="none", extent=(-0.5, 0.5, -0.5, 0.5), cmap=ms_cmap)

    if show_cbar:
        cb = fig.colorbar(im, ax=ax, ticks=[0, 1])
        cb.ax.set_title(rf"$\chi_1$")
        cb.ax.set_yticklabels(["$0$", "$1$"])

    if show_axis:
        ax.axis("on")
        ax.set_xlim(-0.5, 0.5)
        ax.set_ylim(-0.5, 0.5)
        ax.set_xticks([-0.5, 0.0, 0.5])
        ax.set_yticks([-0.5, 0.0, 0.5])
        ax.set_xlabel(r"$\frac{x_1}{l_1}$")
        ax.set_ylabel(r"$\frac{x_2}{l_2}$")
    else:
        ax.axis("off")

    plt.tight_layout()
    if file is not None:
        plt.savefig(file, dpi=300)
    plt.show()


def plot_prec_action(ax, field_ref, init_res, precs, labels, ch_idx=0, plot_res=False, centered=True, ch_label=None):
    idx = (0,0,ch_idx)
    ax_i = 0
    if centered:
        s_abs_max = torch.max(field_ref[idx].cpu().abs())
        s_max, s_min = s_abs_max, -s_abs_max
    else:
        s_max, s_min = torch.max(field_ref[idx].cpu()), torch.min(field_ref[idx].cpu())
    
    prec_actions = []
    with torch.inference_mode():
        for prec in precs:
            prec_action = prec.apply_field(init_res)
            prec_action = prec_action / prec_action[idx].max() * field_ref[idx].abs().max()
            prec_actions.append(prec_action)
            if centered:
                s_abs_max = torch.maximum(s_abs_max, torch.max(prec_action[idx].cpu().abs()))
                s_max, s_min = s_abs_max, -s_abs_max
            else:
                s_max = torch.maximum(s_max, torch.max(prec_action[idx].cpu()))
                s_min = torch.minimum(s_min, torch.min(prec_action[idx].cpu()))

    if plot_res:
        im = ax[0,ax_i].imshow(init_res[idx].cpu().detach(), origin="lower", cmap="seismic")
        if ch_label is None:
            ax[0,ax_i].set_title(r"$\boldsymbol{r}^{(0)}$")
        else:
            ax[0,ax_i].set_title(rf"$\left( \boldsymbol{{r}}^{{(0)}} \right)_{{{ch_label}}}$")
        plt.colorbar(im, ax=ax[ax_i])
        ax_i += 1

    im = ax[0,ax_i].imshow(field_ref[idx].cpu().detach(), origin="lower", cmap="jet", vmax=s_max, vmin=s_min)
    if ch_label is None:
        ax[0,ax_i].set_title(r"$\boldsymbol{u}_{\underline{\mu}}$")
    else:
        ax[0,ax_i].set_title(rf"$\left( \boldsymbol{{u}}_{{\underline{{a}}}} \right)_{{{ch_label}}}$")
    plt.colorbar(im, ax=ax[0,ax_i])
    ax_i += 1
    
    for prec_action, label in zip(prec_actions, labels):
        im = ax[0,ax_i].imshow(prec_action[idx].cpu().detach(), origin="lower", cmap="jet", vmax=s_max, vmin=s_min)
        if ch_label is None:
            ax[0,ax_i].set_title(rf"$\boldsymbol{{s}}_\text{{{label}}}$")
        else:
            ax[0,ax_i].set_title(rf"$\left( \boldsymbol{{s}}_\text{{{label}}} \right)_{{{ch_label}}}$")
        plt.colorbar(im, ax=ax[0,ax_i])
        ax_i += 1

    for ax_handle in ax.ravel():
        ax_handle.set_xticks([])
        ax_handle.set_yticks([])


def plot_convergence(ax, results, labels, colors, metrics, metric_labels, load_names=None, rates=None, show_load_labels=False,
                     xmin=0, xmax=None, bounds=False, ymin=1e-10, ymax=1e0, zoom=False, zoom_it = 10, zoom_tol = 1e-3):
    """
    
    """
    if load_names is None:
        load_names = ["x"]
    if rates is None:
        rates = (None,) * len(results)
    colors = colors[:len(results)]
    while len(labels) < len(results):
        labels.append("CG")

    for load_i, load_name in enumerate(load_names):
        if show_load_labels:
            ax[0, load_i].set_title(f"loading {load_name}")

        for metric_i, (metric, metric_label) in enumerate(zip(metrics, metric_labels)):
            ax[metric_i, load_i].set_xlabel(r"iterations $[-]$")
            ax[metric_i, load_i].set_ylabel(metric_label)
            ax[metric_i, load_i].set_ylim(bottom=ymin, top=ymax)
            if zoom:
                axins = ax[metric_i, load_i].inset_axes([0.2, 0.15, 0.1, 0.45],
                    xlim=(-1, zoom_it), ylim=(zoom_tol, 1e0), xticklabels=[], yticklabels=[])
                axins.grid()
                axins.set_xticks([0, zoom_it], ["$0$", f"${zoom_it}$"])

            for result, label, color, rate in zip(results, labels, colors, rates):
                if result is None:
                    continue
                iters = torch.arange(result["err_history"].shape[0])
                metric_losses = result["losses"][metric]
                if metric_losses.ndim == 1:
                    metric_losses = metric_losses.unsqueeze(-1)
                ax[metric_i, load_i].semilogy(iters, metric_losses[:, load_i].cpu() / metric_losses[0, load_i].cpu(), '-', c=color, label=label)

                if zoom:
                    axins.semilogy(iters, metric_losses[:, load_i].cpu() / metric_losses[0, load_i].cpu(), '-', c=color)

            if zoom:
                ax[metric_i, load_i].indicate_inset_zoom(axins, edgecolor="black")
    
    for ax_handle in ax.ravel():
        ax_handle.grid()
        ax_handle.set_xlim(left=xmin, right=xmax)


def plot_convergence_histogram(ax, results, labels, colors, rtol=1e-6, bins=None, xmin=0, xmax=None, log_scale=False, legend=True):
    """
    
    """
    for result, label, color in zip(results, labels, colors):
        if result is None:
            continue
        metric = get_rel_residual(result)
        iters = (metric > rtol).sum(dim=0)
        ax.hist(iters, edgecolor=color, facecolor=color, bins=bins, label=label)
    if log_scale:
        ax.set_xlim(max(xmin, 1), xmax)
        ax.set_xscale("log")
    else:
        ax.set_xlim(xmin, xmax)
    ax.set_axisbelow(True)
    ax.grid()
    if legend:
        ax.legend()

def get_rel_residual(result):
    """
    
    """
    return result['err_history'].cpu().flatten(start_dim=-2) / result['err_history'].cpu().flatten(start_dim=-2)[0]


def plot_deformed_rve_2d(
    problem,
    disp,
    field,
    loading=None,
    fluctuation_scaling: float = 1.0,
    deformation_scaling: float = 1.0,
    plot_loading: bool = False,
    plot_boundary: bool = False,
    shading: str = "gouraud",
    file: Optional[str] = None,
    vmin: Optional[List[float]] = None,
    vmax: Optional[List[float]] = None,
    figsize: Optional[List[float]] = None
):
    """

    :param disp:
    :param field:
    :param loading:
    :param fluctuation_scaling:
    :param deformation_scaling:
    :param plot_loading:
    :param plot_boundary:
    :param shading:
    :param file:
    :param vmin:
    :param vmax:
    :return:
    """
    def apply_mask(triang, alpha=0.4):
        # Mask triangles with sidelength greater than a threshold alpha
        triangles = triang.triangles
        # Mask off unwanted triangles.
        x = triang.x
        y = triang.y
        xtri = x[triangles] - np.roll(x[triangles], 1, axis=1)
        ytri = y[triangles] - np.roll(y[triangles], 1, axis=1)
        maxi = np.max(np.sqrt(xtri**2 + ytri**2), axis=1)
        # apply masking
        triang.set_mask(maxi > alpha)
    
    if loading is None:
        loading = torch.eye(3, dtype=disp.dtype, device=disp.device)

    n_loadings = loading.shape[0]

    #if problem.n_dim == 2:
    #    disp = disp.transpose(-1, -2)

    deformed_coords = problem.get_deformed_coordinates(
        disp, loading, fluctuation_scaling=fluctuation_scaling, deformation_scaling=deformation_scaling
    )
    loading_coords = problem.get_deformed_coordinates(
        disp, loading, fluctuation_scaling=0.0, deformation_scaling=deformation_scaling
    )
    coords = problem.get_node_coords()
    #boundary_idx = problem.get_boundary_idx()
    #boundary = coords[..., boundary_idx[0], boundary_idx[1]]
    #deformed_boundary = deformed_coords[..., boundary_idx[0], boundary_idx[1]]
    #loading_boundary = loading_coords[..., boundary_idx[0], boundary_idx[1]]

    loadings = [
        rf"$\bar{{\boldsymbol{{\varepsilon}}}}=\bar{{\boldsymbol{{\varepsilon}}}}^{{({i + 1})}}$"
        for i in range(n_loadings)
    ]

    if figsize is None:
        figsize = [6.3, 2.0]

    if problem.n_dim == 3:
        N_cut = problem.n_grid[-1] // 2
        field = field[..., N_cut]
        deformed_coords = deformed_coords[..., N_cut]
            
    fig, ax = plt.subplots(1, n_loadings, figsize=figsize, dpi=300, squeeze=False)
    for load_idx, load_name in enumerate(loadings):
        ax[0, load_idx].axis("off")
        ax[0, load_idx].set_aspect("equal")
        ax[0, load_idx].set_title(load_name)
        tri = Triangulation(deformed_coords[load_idx, 0].ravel(), deformed_coords[load_idx, 1].ravel())
        apply_mask(tri, alpha=0.02)

        if vmin is None:
            vmin_idx = field[load_idx].min().item()
        else:
            vmin_idx = vmin[load_idx]
        if vmax is None:
            vmax_idx = 0.5 * field[load_idx].max().item()
        else:
            vmax_idx = vmax[load_idx]
        
        tpc = ax[0, load_idx].tripcolor(
            tri, field[load_idx].ravel(), cmap="jet", shading=shading, rasterized=True, vmin=vmin_idx, vmax=vmax_idx
        )

        if plot_loading:
            ax[0, load_idx].plot(
                loading_boundary[load_idx, 0].ravel(),
                loading_boundary[load_idx, 1].ravel(),
                "k--",
                lw=1,
            )
        if plot_boundary:
            ax[0, load_idx].plot(
                deformed_boundary[load_idx, 0].ravel(),
                deformed_boundary[load_idx, 1].ravel(),
                "r--",
                lw=1,
            )
        clb = fig.colorbar(tpc)
        clb.ax.set_title(r"$||\boldsymbol{\sigma}|| \,[\mathrm{GPa}]$")
    plt.tight_layout()
    if file is not None:
        plt.savefig(file, dpi=300)
    plt.show()


def plot_deformed_rve_3d(
    problem,
    disp,
    field,
    loadings,
    fluctuation_scaling: float = 1.0,
    deformation_scaling: float = 1.0,
    file: Optional[str] = None,
    vmin: Optional[List[float]] = None,
    vmax: Optional[List[float]] = None,
    figsize: Optional[List[float]] = None
):
    disp = torch.nn.functional.pad(disp, pad=[0,1,0,1,0,1], mode="circular")
    coords = problem.get_node_coords().to(dtype=disp.dtype, device=disp.device)
    deformations = problem.get_deformations(disp.transpose(-1, -3), loadings, fluctuation_scaling=1.0)[0]
    
    x, y, z = coords[2].cpu().numpy(), coords[1].cpu().numpy(), coords[0].cpu().numpy()
    grid = pv.StructuredGrid(x, y, z)
    grid['vectors'] = deformations.flatten(start_dim=1).T.cpu().numpy() * deformation_scaling
    warped = grid.warp_by_vector()
    
    pl = pv.Plotter()
    pl.add_mesh(warped, scalars=field.numpy().ravel(), clim=[vmin, vmax], label="stress norm", cmap="jet", lighting=True, diffuse=0.2, specular=1.0, ambient=0.6, scalar_bar_args={"vertical": True})
    pl.screenshot(filename=file, window_size=figsize)
    pl.show()



from typing import Dict

import torch
from unocg.problems import Problem
from unocg.solvers.torch import CgSolver
from unocg.modules.solvers import CgModule
from unocg.preconditioners import Preconditioner
import timeit


def benchmark(model, inputs, device, n_runs=2, n_warmup=1, verbose=True):
    def run_model():
        with torch.inference_mode():
            model(*inputs)
            torch.cuda.synchronize(device)

    for _ in range(n_warmup):
        run_model()
    model_time = timeit.timeit(run_model, number=n_runs) / n_runs
    if verbose:
        print(f"Runtime per execution: {model_time*1000.:.4f}ms")
    return model_time


def benchmark_cg(cg_module, param_fields, loadings, device, n_runs=2, n_warmup=1, verbose=True):
    with torch.inference_mode():
        guess = cg_module.zero_guess(param_fields, loadings)

        inputs_module = (guess, param_fields, loadings)

        print(f"Overall solver:")
        u = cg_module(*inputs_module)
        time_overall = benchmark(cg_module, inputs_module, device, n_runs=n_runs)

        iter_layer = cg_module.iteration_layers[0]
        print(f"Solver iteration:")
        batch_shape = (*param_fields.shape[:(-cg_module.n_dim - 1)], loadings.shape[0])
        field_shape = (*batch_shape, *cg_module.shape)
        iv_fields, iv_scalars = iter_layer.init_internal_variables(batch_shape=batch_shape, init_residual=u)
        time_iter = benchmark(iter_layer, (u, u, param_fields, loadings, iv_fields, iv_scalars), device, n_runs=n_runs*10)

        print(f"Preconditioner application:")
        time_prec = benchmark(cg_module.prec_model, (u,), n_runs=n_runs*100, n_warmup=n_warmup, device=device);

        print(f"Residual computation:")
        time_matvec = benchmark(cg_module.matvec_model, (u, param_fields), n_runs=n_runs*10, n_warmup=n_warmup, device=device)

        return time_overall, time_iter, time_prec, time_matvec



"""
Thermal problem formulations
"""
import math
import warnings
from typing import Optional, Union, Tuple, Collection, Dict

import torch
import torch.nn.functional as F
from unocg.problems import Problem, BC
from unocg.modules.operators import MatvecModule, RhsModule, FieldModule
from unocg.materials import Material
from unocg.materials.thermal import LinearHeatConduction
from unocg.training.losses.thermal import TempLoss, GradLoss, FluxLoss, ThermalEnergyLoss, HeatCondResidualLoss


class ThermalProblem(Problem):
    """
    Thermal problem
    """

    def __init__(
            self,
            n_grid: Union[torch.Size, Collection[int]],
            material: Material,
            loadings: Optional[torch.Tensor] = None,
            quad_degree: int = 2,
            bc: Optional[BC] = None,
            lengths = None,
    ):
        """
        Initialise the thermal problem.

        :param n_grid:
        :type n_grid: Union[torch.Size, Collection[int]]
        :param material: material model
        :type material: Material
        :param loadings:
        :type loadings: torch.Tensor
        :param quad_degree:
        :type quad_degree: int
        :param bc:
        :type bc: BC
        """
        super().__init__(n_grid=n_grid, bc=bc, quad_degree=quad_degree, lengths=lengths, material=material)
        self._n_channels = 1

        if loadings is None:
            n_loadings = self.n_dim
            self.loadings = torch.eye(n_loadings)
        else:
            assert loadings.ndim == 2
            assert loadings.shape[1] == self.n_dim
            self.loadings = loadings

        if self.bc == BC.PERIODIC:
            self.dof_shape = self.n_grid
        elif self.bc == BC.DIRICHLET:
            self.dof_shape = tuple([n_axis - 1 for n_axis in self.n_grid])
        elif self.bc == BC.DIRICHLET_TB:
            self._n_dof = self.n_channels * (self.n_grid[-1] - 1) * self.n_grid[-2]
            self.dof_shape = (self.n_grid[0] - 1, *self.n_grid[1:])
        elif self.bc == BC.DIRICHLET_LR:
            self._n_dof = self.n_channels * self.n_grid[0] * (self.n_grid[1] - 1)
            self.dof_shape = (*self.n_grid[:1], self.n_grid[1] - 1, *self.n_grid[2:])
        elif self.bc == BC.DIRICHLET_FB:
            self._n_dof = self.n_channels * self.n_grid[0] * (self.n_grid[1] - 1)
            self.dof_shape = (*self.n_grid[:2], self.n_grid[2] - 1, *self.n_grid[3:])
        else:
            raise NotImplementedError()
        self._n_dof = self.n_channels * torch.prod(torch.tensor(self.dof_shape))

        # Precompute kernels for residual computation
        self.grad_operator = self.get_grad_operator()
        self.grad_kernels = self.get_grad_kernels()
        self.div_kernels = self.get_div_kernels()

        self.Idx = None
        self.Lf = None

    def compute_residual(
            self, u: torch.Tensor,
            param_fields: Optional[torch.Tensor] = None,
            loadings: Optional[torch.Tensor] = None,
            state_fields: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Compute residual for the thermal problem.

        :param u:
        :param param_fields:
        :param loadings:
        :param state_fields:
        :return:
        """
        if param_fields is None:
            raise ValueError()
        if loadings is None:
            loadings = self.loadings

        # Move kernels to correct device
        args = {"device": u.device, "dtype": u.dtype}
        self.grad_kernels = self.grad_kernels.to(**args)
        self.div_kernels = self.div_kernels.to(**args)

        temp_grad = self.grad(self.reshape_field(u)) + loadings[(slice(None), None, slice(None)) + self.expand_dims].to(**args)
        flux, _ = self.material_law(temp_grad, param_fields.to(**args))
        r = self.reshape_vec(self.div(flux))

        return r, None

    def matvec(self,
               d: torch.Tensor,
               param_fields: Optional[torch.Tensor] = None,
               state_fields: Optional[torch.Tensor] = None) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Matvec

        :param d:
        :param param_fields:
        :param state_fields:
        :return:
        """
        if param_fields is None:
            raise ValueError()

        # Move kernels to correct device
        args = {"device": d.device, "dtype": d.dtype}
        self.grad_kernels = self.grad_kernels.to(**args)
        self.div_kernels = self.div_kernels.to(**args)

        temp_grad = self.grad(self.reshape_field(d))
        flux, _ = self.material_law(temp_grad, param_fields.to(**args))
        p = self.reshape_vec(self.div(flux))

        return -p, None

    def material_law(self, temp_grad: torch.Tensor, param_fields: Optional[torch.Tensor] = None) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Material law

        :param temp_grad:
        :param param_fields:
        :return:
        """
        if param_fields is None:
            raise ValueError()

        return self.material(temp_grad, param_fields)

    def get_matvec_module(self, dtype, device) -> torch.nn.Module:
        grad_module = self.get_grad_module(dtype=dtype, device=device)
        div_module = self.get_div_module(dtype=dtype, device=device)

        module = MatvecModule(self.n_dim, grad_module, self.material, div_module)
        return module

    def conductivity(self, params: torch.Tensor) -> torch.Tensor:
        """
        Assemble conductivity tensor for one phase

        :param params:
        :return:
        :rtype: torch.Tensor
        """
        if params.ndim == 0:
            params = params.unsqueeze(dim=0)
        assert params.ndim >= 1
        Id = self.Id.to(dtype=params.dtype, device=params.device)
        cond = params[..., None] * Id
        return cond

    def conductivities(self, params: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Assemble conductivity tensor for both phases

        :param params:
        :return:
        :rtype: torch.Tensor
        """
        if params.ndim == 1:
            params = params.unsqueeze(dim=-1)
        return self.conductivity(params[..., 0, :]), self.conductivity(params[..., 1, :])

    def conductivity_field(self, param_fields: torch.Tensor) -> torch.Tensor:
        """
        Assemble isotropic stiffness tensor in mandel notation

        :param param_fields:
        :return:
        :rtype: torch.Tensor
        """
        if param_fields.ndim == 0:
            param_fields = param_fields * torch.ones([1, 1, 1], dtype=param_fields.dtype, device=param_fields.device)
        if param_fields.ndim == 2:
            param_fields = torch.unsqueeze(param_fields, dim=0)
        assert param_fields.ndim >= 3

        Id = self.Id.to(dtype=param_fields.dtype, device=param_fields.device)
        cond = torch.einsum("...i" + self.einsum_dims + ",ij->...ij" + self.einsum_dims, param_fields, Id)
        return cond

    @property
    def Id(self):
        return torch.eye(self.n_dim, self.n_dim)
    
    def construct_sparsity(self):
        n = self.n_grid[-1]
        if self.n_dim == 2:
            # Generate node indices
            x = torch.arange(n)
            y = torch.arange(n)
            X, Y = torch.meshgrid(x, y, indexing="xy")
            idx = torch.arange(n * n).reshape(n, n)

            # Generate element connectivity
            Xe = torch.stack([X, X + 1, X, X + 1], dim=-1)  # x-coordinates of all nodes per element
            Ye = torch.stack([Y, Y, Y + 1, Y + 1], dim=-1)  # y-coordinates of all nodes per element
            idx_pad = F.pad(idx.unsqueeze(0), [0, 1, 0, 1], mode="circular").squeeze()
            conn = idx_pad[Ye, Xe]

            # Generate indices for global stiffness matrix
            idx_x = conn.repeat(1, 1, 4)
            idx_y = conn.repeat_interleave(4, dim=-1)
            self.Idx = torch.stack([idx_x.ravel(), idx_y.ravel()])

            # Determine indices belonging to free DOFs
            if self.bc == BC.PERIODIC:
                self.idx_free = idx.ravel()
            elif self.bc == BC.DIRICHLET:
                self.idx_free = idx[1:, 1:].ravel()
            elif self.bc == BC.DIRICHLET_LR:
                self.idx_free = idx[1:, :].ravel()
            elif self.bc == BC.DIRICHLET_TB:
                self.idx_free = idx[:, 1:].ravel()
            else:
                raise NotImplementedError()
            
            # Assemble scatter matrix for free DOFs
            self.Lf = self.get_scatter_matrix(self.idx_free, shape=(self.idx_free.shape[0], n * n))
        
        elif self.n_dim == 3:
            # Generate node indices
            x = torch.arange(n)
            y = torch.arange(n)
            z = torch.arange(n)
            X, Y, Z = torch.meshgrid(x, y, z, indexing="ij")  # TODO: check indexing
            idx = torch.arange(n * n * n).reshape(n, n, n)

            # Generate element connectivity
            Xe = torch.stack([X, X + 1, X, X + 1, X, X + 1, X, X + 1], dim=-1)  # x-coordinates of all nodes per element, TODO: change
            Ye = torch.stack([Y, Y, Y + 1, Y + 1, Y, Y, Y + 1, Y + 1], dim=-1)  # y-coordinates of all nodes per element, TODO: change
            Ze = torch.stack([Z, Z, Z, Z, Z + 1, Z + 1, Z + 1, Z + 1], dim=-1)  # z-coordinates of all nodes per element, TODO: change
            idx_pad = F.pad(idx.unsqueeze(0), [0, 1, 0, 1, 0, 1], mode="circular").squeeze()
            conn = idx_pad[Xe, Ye, Ze]

            # Generate indices for global stiffness matrix
            idx_x = conn.repeat(1, 1, 1, 8)
            idx_y = conn.repeat_interleave(8, dim=-1)
            self.Idx = torch.stack([idx_x.ravel(), idx_y.ravel()])

            # Determine indices belonging to free DOFs
            if self.bc == BC.PERIODIC:
                self.idx_free = idx.ravel()
            elif self.bc == BC.DIRICHLET:
                self.idx_free = idx[1:, 1:, 1:].ravel()
            elif self.bc == BC.DIRICHLET_LR:
                self.idx_free = idx[1:, :, :].ravel()
            elif self.bc == BC.DIRICHLET_TB:
                self.idx_free = idx[:, 1:, :].ravel()
            else:
                raise NotImplementedError()
            
            # Assemble scatter matrix for free DOFs
            self.Lf = self.get_scatter_matrix(self.idx_free, shape=(self.idx_free.shape[0], n * n * n))
            
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")


    def assemble_matrix(self, param_fields: torch.Tensor, coalesce=True) -> torch.Tensor:
        """
        Assemble global stiffness matrix

        :param param_fields:
        :return:
        """
        if (self.Idx is None) or (self.Lf is None):
            self.construct_sparsity()
        self.Idx = self.Idx.to(dtype=param_fields.dtype, device=param_fields.device)
        self.Lf = self.Lf.to(dtype=param_fields.dtype, device=param_fields.device)
        
        if self.n_dim == 2:
            # Generate node indices
            #param_field = param_fields.transpose(-1, -2)
            n = param_fields.shape[-1]

            # Assemble element stiffness matrices
            cond_field = self.conductivity_field(param_fields)
            ke = torch.flatten(self.get_ke(cond_field), start_dim=-3)

            # Assemble global stiffness matrix
            stiffness = torch.sparse_coo_tensor(self.Idx, ke.ravel(), (n * n, n * n))
        
        elif self.n_dim == 3:
            # Generate node indices
            # param_field = param_fields.transpose(-1, -2)  # TODO: change
            n = param_fields.shape[-1]

            # Assemble element stiffness matrices
            cond_field = self.conductivity_field(param_fields)
            ke = torch.flatten(self.get_ke(cond_field), start_dim=-3)

            # Assemble global stiffness matrix
            stiffness = torch.sparse_coo_tensor(self.Idx, ke.ravel(), (n * n * n, n * n * n))
            
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")
        
        if coalesce:
            stiffness = stiffness.coalesce()

        # Extract stiffness matrix acting on free DOFs
        if self.bc == BC.PERIODIC:
            stiffness_free = stiffness
        else:
            with warnings.catch_warnings():
                # PyTorch UserWarning for sparse matrices can be ignored
                warnings.filterwarnings("ignore")
                stiffness_free = self.Lf @ stiffness @ self.Lf.T

        return stiffness_free

    def assemble_matrix_diag(self, image, params):
        # TODO: correct interface contributions
        cond0, cond1 = self.conductivity(params[...,0,:]), self.conductivity(params[...,1,:])
        for _ in range(self.n_dim):
            cond0, cond1 = cond0.unsqueeze(-1), cond1.unsqueeze(-1)
        ke0, ke1 = self.get_ke(cond0), self.get_ke(cond1)
        A_diag0, A_diag1 = torch.diag(ke0.squeeze()).sum(), torch.diag(ke1.squeeze()).sum()
        A_diag_approx = A_diag0 * (1. - image.ravel()) + A_diag1 * image.ravel()
        return A_diag_approx

    def assemble_rhs(self, param_fields: torch.Tensor, loadings: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        Assemble right hand side vector of linear system for given loading

        Equivalent to residual computation with zero guess for temperature field

        :param param_fields:
        :param loadings:
        :return:
        """
        if param_fields is None:
            raise ValueError()
        if loadings is None:
            loadings = self.loadings

        # Move kernels to correct device
        args = {"device": param_fields.device, "dtype": param_fields.dtype}
        self.grad_kernels = self.grad_kernels.to(**args)
        self.div_kernels = self.div_kernels.to(**args)

        batch_shape = param_fields.shape[:self.ch_dim]
        grad_shape = (*batch_shape, loadings.shape[0], self.n_gauss, loadings.shape[-1], *self.n_grid)
        grad = loadings[(slice(None), None, slice(None)) + self.expand_dims].expand(grad_shape).to(**args)
        flux, _ = self.material_law(grad, param_fields)
        rhs = self.reshape_vec(self.div(flux))
        return rhs
    
    def get_rhs_module(self, dtype, device) -> torch.nn.Module:
        div_module = self.get_div_module(dtype=dtype, device=device)

        module = RhsModule(
            n_dim=self.n_dim,
            shape=(self.n_gauss, self.n_dim, *self.n_grid),
            model=self.material,
            div_module=div_module,
        )
        return module

    @property
    def n_gauss(self) -> int:
        return self.div_kernels.shape[-(3 + self.n_dim)]

    def compute_field(
            self,
            u: torch.Tensor,
            param_fields: Optional[torch.Tensor] = None,
            loadings: Optional[torch.Tensor] = None,
            state_fields: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """

        :param u:
        :param param_fields:
        :param loadings:
        :param state_fields:
        :return:
        """
        if param_fields is None:
            raise ValueError()

        temp = self.reshape_field(u)
        flux = self.compute_flux(temp, param_fields, loadings=loadings, reduce=True)

        if self.n_dim == 2:
            if self.bc == BC.DIRICHLET:
                temp = F.pad(temp, [1, 0, 1, 0], mode="constant", value=0.0)
            elif self.bc == BC.DIRICHLET_TB:
                temp = F.pad(temp, [1, 0, 0, 0], mode="constant", value=0.0)
            elif self.bc == BC.DIRICHLET_LR:
                temp = F.pad(temp, [0, 0, 1, 0], mode="constant", value=0.0)
        elif self.n_dim == 3:
            if self.bc == BC.DIRICHLET:
                temp = F.pad(temp, [1, 0, 1, 0, 1, 0], mode="constant", value=0.0)
            elif self.bc == BC.DIRICHLET_TB:
                temp = F.pad(temp, [1, 0, 0, 0, 0, 0], mode="constant", value=0.0)
            elif self.bc == BC.DIRICHLET_LR:
                temp = F.pad(temp, [0, 0, 1, 0, 0, 0], mode="constant", value=0.0)
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} not supported for this operation")

        field = torch.cat([temp, flux], dim=self.ch_dim)
        return field
    
    def get_field_module(self, dtype, device):
        padding_layers = []
        if self.n_dim == 2:
            if self.bc == BC.DIRICHLET:
                padding_layers = [torch.nn.ZeroPad2d((1, 0, 1, 0))]
            elif self.bc == BC.DIRICHLET_TB:
                padding_layers = [torch.nn.ZeroPad2d((1, 0, 0, 0))]
            elif self.bc == BC.DIRICHLET_LR:
                padding_layers = [torch.nn.ZeroPad2d((0, 0, 1, 0))]
        elif self.n_dim == 3:
            if self.bc == BC.DIRICHLET:
                padding_layers = [torch.nn.ZeroPad3d((1, 0, 1, 0, 1, 0))]
            elif self.bc == BC.DIRICHLET_TB:
                padding_layers = [torch.nn.ZeroPad3d((1, 0, 0, 0, 0, 0))]
            elif self.bc == BC.DIRICHLET_TB:
                padding_layers = [torch.nn.ZeroPad3d((0, 0, 1, 0, 0, 0))]
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} not supported for this operation")

        return FieldModule(
            ch_dim=self.ch_dim,
            expand_dims=self.expand_dims,
            grad_module=self.get_grad_module(dtype=dtype, device=device),
            model=self.material,
            padding_layers=padding_layers,
        )
    
    def get_hom_response(
            self,
            u: torch.Tensor,
            param_fields: Optional[torch.Tensor] = None,
            loadings: Optional[torch.Tensor] = None,
            state_fields: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Compute homogenized response based on DOF vector u

        :param u:
        :param param_fields:
        :param loadings:
        :param state_fields:
        :return:
        """
        field = self.compute_field(u, param_fields, loadings, state_fields)
        hom_response = field[(Ellipsis, slice(1, None)) + self.dims].mean(self.dims_list)
        return hom_response

    def compute_flux(
            self,
            temp: torch.Tensor,
            param_fields: torch.Tensor,
            loadings: Optional[torch.Tensor] = None,
            reduce: bool = False,
    ) -> torch.Tensor:
        """
        Compute temperature gradient field based on the temperature field for a given

        :param temp:
        :param param_fields:
        :param loadings:
        :param reduce: reduce flux at all gauss points to mean value for each element
        :return: flux
        """
        if temp.ndim == 3:
            temp = torch.unsqueeze(temp, dim=0)
        assert temp.ndim >= 4
        assert temp.shape[self.ch_dim] == 1
        n_loadings = temp.shape[self.ch_dim - 1]
        if loadings is None:
            loadings = torch.eye(n_loadings, dtype=temp.dtype, device=temp.device)
        else:
            n_loadings = temp.shape[self.ch_dim - 1] if temp.ndim >= 4 else 1
        assert loadings.shape[0] == n_loadings
        assert loadings.shape[1] == self.n_dim

        temp_grad_fluctuations = self.grad(temp)
        temp_grad = temp_grad_fluctuations + loadings[(slice(None), None, slice(None)) + self.expand_dims]
        flux, _ = self.material_law(temp_grad, param_fields)

        if reduce:
            flux = flux.mean(-(2 + self.n_dim))  # mean value over all gauss points of each element
        return flux

    def grad(self, temp: torch.Tensor) -> torch.Tensor:
        """
        Compute the symmetric gradient of a periodic 2d vector field disp

        :param temp:
        :return:
        """
        if temp.ndim == self.n_dim + 1:
            temp = torch.unsqueeze(temp, dim=0)
        assert temp.ndim >= self.n_dim + 2
        # assert temp.shape[-1] == temp.shape[-2]
        assert temp.shape[self.ch_dim] == self.n_channels

        if self.n_dim == 2:
            return self.grad_2d(temp)
        elif self.n_dim == 3:
            return self.grad_3d(temp)
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")

    def grad_2d(self, temp: torch.Tensor) -> torch.Tensor:
        """
        Compute the 2D gradient of a periodic vector field temp

        :param temp:
        :return:
        """
        # Construct kernels if not given
        kernels = self.grad_kernels.to(device=temp.device, dtype=temp.dtype)

        # Determine appropriate padding
        if self.bc == BC.PERIODIC:
            periodic_padding = [0, 1, 0, 1]
            constant_padding = [0, 0, 0, 0]
        elif self.bc == BC.DIRICHLET:
            periodic_padding = [0, 0, 0, 0]
            constant_padding = [1, 1, 1, 1]
        elif self.bc == BC.DIRICHLET_LR:
            periodic_padding = [0, 1, 0, 0]
            constant_padding = [0, 0, 1, 1]
        elif self.bc == BC.DIRICHLET_TB:
            periodic_padding = [0, 0, 0, 1]
            constant_padding = [1, 1, 0, 0]
        else:
            raise NotImplementedError()

        # Prepare reshaping as F.pad and F.conv2d only accept 3D and 4D tensors
        original_shape = torch.tensor(temp.shape)
        output_shape = torch.cat([original_shape[:self.ch_dim], torch.tensor(kernels.shape[:1]), original_shape[self.ch_dim:]])
        output_shape[self.ch_dim] = self.n_dim

        # Periodic padding of the temperature field to treat boundaries correctly
        temp_pad = temp
        if sum(periodic_padding) > 0:
            temp_pad = F.pad(temp_pad.flatten(end_dim=-3), periodic_padding, mode="circular").reshape(
                -1,
                original_shape[-3],
                temp_pad.shape[-2] + periodic_padding[2] + periodic_padding[3],
                temp_pad.shape[-1] + periodic_padding[0] + periodic_padding[1],
            )
        # Constant padding of the displacement field to treat boundaries correctly
        if sum(constant_padding) > 0:
            temp_pad = F.pad(temp_pad.flatten(end_dim=-3), constant_padding, mode="constant", value=0.0).reshape(
                -1,
                original_shape[-3],
                temp_pad.shape[-2] + constant_padding[2] + constant_padding[3],
                temp_pad.shape[-1] + constant_padding[0] + constant_padding[1],
            )

        # Compute temperature gradient tensor using a 2d convolution
        g = F.conv2d(temp_pad, kernels.flatten(end_dim=-4))
        return g.reshape(*temp.shape[:-3], kernels.shape[0], 2, *g.shape[-2:])

    def grad_3d(self, temp: torch.Tensor) -> torch.Tensor:
        """
        Compute the 3D gradient of a periodic vector field temp

        :param temp:
        :return:
        """
        # Construct kernels if not given
        kernels = self.grad_kernels.to(device=temp.device, dtype=temp.dtype)

        # Determine appropriate padding
        if self.bc == BC.PERIODIC:
            periodic_padding = [0, 1, 0, 1, 0, 1]
            constant_padding = [0, 0, 0, 0, 0, 0]
        elif self.bc == BC.DIRICHLET:
            periodic_padding = [0, 0, 0, 0, 0, 0]
            constant_padding = [1, 1, 1, 1, 1, 1]
        elif self.bc == BC.DIRICHLET_LR:
            periodic_padding = [0, 1, 0, 0, 0, 1]
            constant_padding = [0, 0, 1, 1, 0, 0]
        elif self.bc == BC.DIRICHLET_TB:
            periodic_padding = [0, 0, 0, 1, 0, 1]
            constant_padding = [1, 1, 0, 0, 0, 0]
        else:
            raise NotImplementedError()

        # Prepare reshaping as F.pad and F.conv2d only accept 3D and 4D tensors
        original_shape = torch.tensor(temp.shape)
        output_shape = torch.cat([original_shape[:self.ch_dim], torch.tensor(kernels.shape[:1]), original_shape[self.ch_dim:]])
        output_shape[self.ch_dim] = self.n_dim

        # Periodic padding of the temperature field to treat boundaries correctly
        temp_pad = temp
        if sum(periodic_padding) > 0:
            temp_pad = F.pad(temp_pad.flatten(end_dim=-4), periodic_padding, mode="circular").reshape(
                -1,
                1,
                temp_pad.shape[-3] + periodic_padding[4] + periodic_padding[5],
                temp_pad.shape[-2] + periodic_padding[2] + periodic_padding[3],
                temp_pad.shape[-1] + periodic_padding[0] + periodic_padding[1],
            )
        # Constant padding of the displacement field to treat boundaries correctly
        if sum(constant_padding) > 0:
            temp_pad = F.pad(temp_pad.flatten(end_dim=-4), constant_padding, mode="constant", value=0.0).reshape(
                -1,
                1,
                temp_pad.shape[-3] + constant_padding[4] + constant_padding[5],
                temp_pad.shape[-2] + constant_padding[2] + constant_padding[3],
                temp_pad.shape[-1] + constant_padding[0] + constant_padding[1],
            )
        output_shape[-1] = temp_pad.shape[-1] - 1
        output_shape[-2] = temp_pad.shape[-2] - 1

        # Compute temperature gradient tensor using a 2d convolution
        g = F.conv3d(temp_pad, kernels.flatten(end_dim=-5))
        return g.reshape(*temp.shape[:-4], kernels.shape[0], self.n_dim, *g.shape[-3:])

    def get_grad_module(self, dtype, device, unflatten=True) -> torch.nn.Module:
        grad_kernels = self.grad_kernels.to(dtype=dtype, device=device)
        if self.n_dim == 2:
            conv_grad = torch.nn.Conv2d(self.n_channels, self.n_dim * self.n_gauss, grad_kernels.shape[-self.n_dim:],
                                        bias=False, device=device, dtype=dtype)
            conv_grad.weight = torch.nn.Parameter(grad_kernels.flatten(end_dim=-(2 + self.n_dim)), requires_grad=False)

            if self.bc == BC.PERIODIC:
                padding_layers = [torch.nn.CircularPad2d((0, 1) * self.n_dim)]
            elif self.bc == BC.DIRICHLET:
                padding_layers = [torch.nn.ZeroPad2d((1, 1) * self.n_dim)]
            elif self.bc == BC.DIRICHLET_TB:
                padding_layers = [torch.nn.CircularPad2d((0, 0, 0, 1)), torch.nn.ZeroPad2d((1, 1, 0, 0))]
            elif self.bc == BC.DIRICHLET_LR:
                padding_layers = [torch.nn.CircularPad2d((0, 1, 0, 0)), torch.nn.ZeroPad2d((0, 0, 1, 1))]
            else:
                raise NotImplementedError(f"Boundary condition {self.bc} is not supported")
            
        elif self.n_dim == 3:
            conv_grad = torch.nn.Conv3d(self.n_channels, self.n_dim * self.n_gauss, grad_kernels.shape[-self.n_dim:],
                                        bias=False, device=device, dtype=dtype)
            conv_grad.weight = torch.nn.Parameter(grad_kernels.flatten(end_dim=-(2 + self.n_dim)), requires_grad=False)

            if self.bc == BC.PERIODIC:
                padding_layers = [torch.nn.CircularPad3d((0, 1) * self.n_dim)]
            elif self.bc == BC.DIRICHLET:
                padding_layers = [torch.nn.ZeroPad3d((1, 1) * self.n_dim)]
            elif self.bc == BC.DIRICHLET_TB:
                padding_layers = [torch.nn.CircularPad3d((0, 0, 0, 1, 0, 1)), torch.nn.ZeroPad3d((1, 1, 0, 0, 0, 0))]
            elif self.bc == BC.DIRICHLET_LR:
                padding_layers = [torch.nn.CircularPad3d((0, 1, 0, 1, 0, 0)), torch.nn.ZeroPad3d((0, 0, 0, 0, 1, 1))]
            else:
                raise NotImplementedError(f"Boundary condition {self.bc} is not supported")
            
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")

        model = torch.nn.Sequential(
            *padding_layers,
            conv_grad,
            torch.nn.Unflatten(1, (self.n_gauss, self.n_dim))
        )
        return model

    def div(self, flux: torch.Tensor) -> torch.Tensor:
        """
        Compute the divergence of a periodic vector field flux

        :param flux:
        :return:
        """
        if flux.ndim == self.n_dim + 1:
            flux = torch.unsqueeze(flux, dim=0)
        assert flux.ndim >= self.n_dim + 2
        assert flux.shape[self.ch_dim] == self.n_dim

        if self.n_dim == 2:
            return self.div_2d(flux)
        elif self.n_dim == 3:
            return self.div_3d(flux)
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")

    def div_2d(self, flux: torch.Tensor) -> torch.Tensor:
        """
        Compute the 2D divergence of a periodic vector field flux

        :param flux:
        :return:
        """
        # Construct kernels if not given
        kernels = self.div_kernels.to(device=flux.device, dtype=flux.dtype)

        # Determine appropriate padding
        if self.bc == BC.PERIODIC:
            periodic_padding = [1, 0, 1, 0]
            constant_padding = [0, 0, 0, 0]
        elif self.bc == BC.DIRICHLET:
            periodic_padding = [0, 0, 0, 0]
            constant_padding = [0, 0, 0, 0]
        elif self.bc == BC.DIRICHLET_LR:
            periodic_padding = [1, 0, 0, 0]
            constant_padding = [0, 0, 0, 0]
        elif self.bc == BC.DIRICHLET_TB:
            periodic_padding = [0, 0, 1, 0]
            constant_padding = [0, 0, 0, 0]
        else:
            raise NotImplementedError()

        # Prepare reshaping as F.pad and F.conv2d only accept 3D and 4D tensors
        # Original shape of flux is (..., L, G, 2, X, Y)
        # Output shape of res will be (..., L, 1, X, Y)
        original_shape = torch.tensor(flux.shape)
        output_shape = torch.cat([original_shape[:-4], original_shape[-3:]])
        output_shape[-3] = 1

        # Padding of the flux field to treat boundaries correctly
        # Reshape flux before padding to (-1, (G * 2), X, Y)
        flux_reshaped = flux.flatten(end_dim=-5).flatten(start_dim=-4, end_dim=-3)
        # Shape of flux_pad after padding is (-1, (G * 2), X + 1, Y + 1)
        # Periodic padding
        flux_pad = flux_reshaped
        if sum(periodic_padding) > 0:
            flux_pad = F.pad(flux_pad, periodic_padding, mode="circular")
        # Constant padding
        if sum(constant_padding) > 0:
            flux_pad = F.pad(flux_pad, constant_padding, mode="constant", value=0.0)

        # Compute residual field using a 2d convolution
        vol_e = 1.0 / torch.tensor(flux.shape[-2:]).prod()
        res = F.conv2d(flux_pad, kernels.transpose(0, 1).flatten(start_dim=1, end_dim=2)) * vol_e / kernels.shape[0]
        output_shape[-1] = flux_pad.shape[-1] - 1
        output_shape[-2] = flux_pad.shape[-2] - 1
        return res.reshape(*flux.shape[:-4], 1, *res.shape[-2:])

    def div_3d(self, flux: torch.Tensor) -> torch.Tensor:
        """
        Compute the 3D divergence of a periodic vector field flux

        :param flux:
        :return:
        """
        # Construct kernels if not given
        kernels = self.div_kernels.to(device=flux.device, dtype=flux.dtype)

        # Determine appropriate padding
        if self.bc == BC.PERIODIC:
            periodic_padding = [1, 0, 1, 0, 1, 0]
            constant_padding = [0, 0, 0, 0, 0, 0]
        elif self.bc == BC.DIRICHLET:
            periodic_padding = [0, 0, 0, 0, 0, 0]
            constant_padding = [0, 0, 0, 0, 0, 0]
        elif self.bc == BC.DIRICHLET_LR:
            periodic_padding = [1, 0, 0, 0, 1, 0]
            constant_padding = [0, 0, 0, 0, 0, 0]
        elif self.bc == BC.DIRICHLET_TB:
            periodic_padding = [0, 0, 1, 0, 1, 0]
            constant_padding = [0, 0, 0, 0, 0, 0]
        else:
            raise NotImplementedError()

        # Prepare reshaping as F.pad and F.conv2d only accept 3D and 4D tensors
        # Original shape of flux is (..., L, G, 3, X, Y, Z)
        # Output shape of res will be (..., L, 1, X, Y, Z)
        original_shape = torch.tensor(flux.shape)
        output_shape = torch.cat([original_shape[:(self.ch_dim - 1)], original_shape[self.ch_dim:]])
        output_shape[self.ch_dim] = 1

        # Padding of the stress field to treat boundaries correctly
        # Reshape flux before padding to (-1, (G * 6), X, Y, Z)
        flux_reshaped = flux.flatten(end_dim=-6).flatten(start_dim=self.ch_dim - 1, end_dim=self.ch_dim)
        # Shape of flux_pad after padding is (-1, (G * 6), X + 1, Y + 1, Z + 1)
        # Periodic padding
        flux_pad = flux_reshaped
        if sum(periodic_padding) > 0:
            flux_pad = F.pad(flux_pad, periodic_padding, mode="circular")
        # Constant padding
        if sum(constant_padding) > 0:
            flux_pad = F.pad(flux_pad, constant_padding, mode="constant", value=0.0)

        # Compute residual field using a 3d convolution
        vol_e = 1.0 / torch.tensor(flux.shape[-3:]).prod()
        res = F.conv3d(flux_pad, kernels.transpose(0, 1).flatten(start_dim=1, end_dim=2)) * vol_e / kernels.shape[0]
        output_shape[-1] = flux_pad.shape[-1] - 1
        output_shape[-2] = flux_pad.shape[-2] - 1
        return res.reshape(*flux.shape[:-5], 1, *res.shape[-3:])

    def get_div_module(self, dtype, device, reduced=False) -> torch.nn.Module:
        div_kernels = self.div_kernels.to(dtype=dtype, device=device)
        if reduced:
            div_kernels = div_kernels.mean(0).unsqueeze(0)

        if self.n_dim == 2:
            Conv = torch.nn.Conv2d
            conv_div = Conv(self.n_dim * self.n_gauss, 1, div_kernels.shape[-self.n_dim:],
                            bias=False, device=device, dtype=dtype)
            vol_e = 1.0 / torch.tensor(self.n_grid).prod()
            conv_div_weight = div_kernels.transpose(0, 1).flatten(start_dim=1, end_dim=2) * vol_e / div_kernels.shape[0]
            conv_div.weight = torch.nn.Parameter(conv_div_weight, requires_grad=True)

            if self.bc == BC.PERIODIC:
                padding_layers = [torch.nn.CircularPad2d((1, 0, 1, 0))]
            elif self.bc == BC.DIRICHLET:
                padding_layers = []
            elif self.bc == BC.DIRICHLET_TB:
                padding_layers = [torch.nn.CircularPad2d((0, 0, 1, 0))]
            elif self.bc == BC.DIRICHLET_LR:
                padding_layers = [torch.nn.CircularPad2d((1, 0, 0, 0))]
            else:
                raise NotImplementedError(f"Boundary condition {self.bc} is not supported")
        elif self.n_dim == 3:
            # Define div Conv3D
            conv_div = torch.nn.Conv3d(3 * self.n_gauss, 1, div_kernels.shape[-self.n_dim:],
                                       bias=False, device=device, dtype=dtype)
            vol_e = 1.0 / torch.tensor(self.n_grid).prod()
            conv_div_weight = div_kernels.transpose(0, 1).flatten(start_dim=1, end_dim=2) * vol_e / div_kernels.shape[0]
            conv_div.weight = torch.nn.Parameter(conv_div_weight, requires_grad=True)

            if self.bc == BC.PERIODIC:
                padding_layers = [torch.nn.CircularPad3d((1, 0, 1, 0, 1, 0))]
            elif self.bc == BC.DIRICHLET:
                padding_layers = []
            elif self.bc == BC.DIRICHLET_TB:
                padding_layers = [torch.nn.CircularPad3d((0, 0, 1, 0, 1, 0))]
            elif self.bc == BC.DIRICHLET_LR:
                padding_layers = [torch.nn.CircularPad3d((1, 0, 0, 0, 1, 0))]
            else:
                raise NotImplementedError(f"Boundary condition {self.bc} is not supported")
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")
        
        model = torch.nn.Sequential(
            torch.nn.Flatten(start_dim=1, end_dim=2),
            *padding_layers,
            conv_div,
        )
        return model

    def get_ke(self, cond_field: torch.Tensor) -> torch.Tensor:
        """
        Get element stiffness

        :param cond_field:
        :return: element stiffness tensor with shape batch_size x *n_grid x 4 x 4
        """
        while cond_field.ndim < 2 + self.n_dim:
            cond_field = cond_field.unsqueeze(-1)

        assert cond_field.ndim >= 4
        assert cond_field.shape[-1] == cond_field.shape[-2]

        args = {"dtype": cond_field.dtype, "device": cond_field.device}
        vol_e = 1. / torch.tensor(self.n_grid).prod()
        b = self.get_grad_operator().to(**args)
        bt = b.transpose(-1, -2)

        # Loop over Gauss points
        ke = torch.einsum("ij,...jk" + self.einsum_dims + ",kl->..." + self.einsum_dims + "il",
                               bt[0], cond_field, b[0]) * vol_e / self.n_gauss
        for g in range(1, self.n_gauss):
            ke += torch.einsum("ij,...jk" + self.einsum_dims + ",kl->..." + self.einsum_dims + "il",
                               bt[g], cond_field, b[g]) * vol_e / self.n_gauss
        return ke

    def compute_fourier_kernel(self, cond: torch.Tensor, full=False) -> torch.Tensor:
        """

        :param cond:
        :return:
        """
        if self.n_dim == 2:
            return self.compute_fourier_kernel_2d(cond, full=full)
        elif self.n_dim == 3:
            return self.compute_fourier_kernel_3d(cond, full=full)
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")

    def compute_fourier_kernel_2d(self, cond: Optional[torch.Tensor] = None, full=False) -> torch.Tensor:
        """

        :param cond:
        :return:
        """
        ny, nx = self.n_grid
        nh = nx if full else nx // 2 + 1

        complex_type = self.get_complex_dtype(cond.dtype)
        range_nx = torch.arange(nx, device=cond.device).type(complex_type)
        range_ny = torch.arange(ny, device=cond.device).type(complex_type)
        xi1 = torch.exp(2 * math.pi * 1j * 2 * range_nx / (2 * nx))[:nh]
        xi2 = torch.exp(2 * math.pi * 1j * 2 * range_ny / (2 * ny))
        ones_x = torch.ones(nh, dtype=cond.dtype, device=cond.device)
        ones_y = torch.ones(ny, dtype=cond.dtype, device=cond.device)

        A = torch.zeros((4, ny, nh), dtype=complex_type, device=cond.device)
        A[0] = ones_x[None, :] * ones_y[:, None]
        A[1] = xi1[None, :] * ones_y[:, None]
        A[2] = ones_x[None, :] * xi2[:, None]
        A[3] = xi1[None, :] * xi2[:, None]
        AA = torch.einsum("iyx,jyx->ijyx", A, A.conj())

        ke = self.get_ke(cond).squeeze().to(dtype=AA.dtype, device=AA.device)
        fk = torch.einsum("ij,ijyx->yx", ke, AA).real
        return fk

    def compute_fourier_kernel_3d(self, cond: Optional[torch.Tensor] = None, full=False) -> torch.Tensor:
        """
        Compute the Fourier kernel of the homogeneous stiffness matrix related to cond in 3D

        :param cond: conductivity tensor for the homogeneous stiffness matrix
        :return: Fourier kernel
        """
        nz, ny, nx = self.n_grid
        nh = nx if full else nx // 2 + 1

        # Construct frequency variables in all 3 dimensions
        complex_type = self.get_complex_dtype(cond.dtype)
        range_nx = torch.arange(nx, device=cond.device).type(complex_type)
        range_ny = torch.arange(ny, device=cond.device).type(complex_type)
        range_nz = torch.arange(nz, device=cond.device).type(complex_type)
        xi1 = torch.exp(2.0 * math.pi * 1j * range_nx / nx)[:nh]
        xi2 = torch.exp(2.0 * math.pi * 1j * range_ny / ny)
        xi3 = torch.exp(2.0 * math.pi * 1j * range_nz / nz)
        ones_x = torch.ones(nh, dtype=cond.dtype, device=cond.device)
        ones_y = torch.ones(ny, dtype=cond.dtype, device=cond.device)
        ones_z = torch.ones(nz, dtype=cond.dtype, device=cond.device)

        # Assemble auxiliary vector field with 8 components on the 3D domain for the Fourier kernel
        # This vector field can be interpreted as Fourier representation of the FEM grad/B operator
        A = torch.zeros((8, nz, ny, nh), dtype=complex_type, device=cond.device)
        A[0] = ones_x[None, None, :] * ones_y[None, :, None] * ones_z[:, None, None]
        A[1] = xi1[None, None, :] * ones_y[None, :, None] * ones_z[:, None, None]
        A[2] = ones_x[None, None, :] * xi2[None, :, None] * ones_z[:, None, None]
        A[3] = xi1[None, None, :] * xi2[None, :, None] * ones_z[:, None, None]
        A[4] = ones_x[None, None, :] * ones_y[None, :, None] * xi3[:, None, None]
        A[5] = xi1[None, None, :] * ones_y[None, :, None] * xi3[:, None, None]
        A[6] = ones_x[None, None, :] * xi2[None, :, None] * xi3[:, None, None]
        A[7] = xi1[None, None, :] * xi2[None, :, None] * xi3[:, None, None]

        # Compute Fourier kernel based on the element stiffness matrix ke and the auxiliary vector field A
        AA = torch.einsum("i" + self.einsum_dims + ",j" + self.einsum_dims + "->ij" + self.einsum_dims, A, A.conj())
        ke = self.get_ke(cond).squeeze().to(dtype=AA.dtype, device=AA.device)
        fk = torch.einsum("ij,ij" + self.einsum_dims + "->" + self.einsum_dims, ke, AA).real
        return fk

    def compute_fundamental_solution(self, cond: torch.Tensor, full=False) -> torch.Tensor:
        """
        Compute the FANS fundamental solution using a Fourier kernel for a given conductivity tensor cond

        :param cond: conductivity tensor for the homogeneous stiffness matrix
        :return: FANS fundamental solution
        """
        fk = self.compute_fourier_kernel(cond, full=full)

        tol = 1e-14
        phi_hat = torch.where(fk < tol, 0.0, 1.0 / fk)
        return phi_hat

    def get_grad_operator(self) -> torch.Tensor:
        """
        Get FEM gradient operator

        :return:
        """
        # TODO: more general, so far only 2D
        rule1d, rule1dw = self.get_quad_rule()

        n_gauss_dir = len(rule1d)
        n_gauss = n_gauss_dir ** self.n_dim
        n_nodes = 2 ** self.n_dim
        h = torch.ones(self.n_dim, dtype=self.dtype) / torch.tensor(self.n_grid, dtype=self.dtype)

        grad_operators = torch.zeros((n_gauss, self.n_dim, self.n_channels * n_nodes), dtype=self.dtype)

        if self.n_dim == 2:
            qw = torch.outer(rule1dw, rule1dw).ravel()  # weights in 2D
            qx = rule1d.expand((n_gauss_dir, n_gauss_dir)).T.ravel()
            qy = rule1d.expand((n_gauss_dir, n_gauss_dir)).ravel()

            B = lambda xi, eta: torch.stack(
                [
                    torch.stack([eta - 1.0, -(eta - 1.0), -eta, eta], dim=-1),
                    torch.stack([xi - 1.0, -xi, 1.0 - xi, xi], dim=-1),
                ],
                dim=-1,
            ).transpose(-1, -2)
            grad_operators = B(qx, qy) / h[None, :, None]
        elif self.n_dim == 3:
            qw = (rule1dw[:, None, None] * rule1dw[None, :, None] * rule1dw[None, None, :]).ravel()
            qx = rule1d.expand(len(rule1d), len(rule1d), len(rule1d)).permute(0, 1, 2).ravel()
            qy = rule1d.expand(len(rule1d), len(rule1d), len(rule1d)).permute(0, 2, 1).ravel()
            qz = rule1d.expand(len(rule1d), len(rule1d), len(rule1d)).permute(2, 1, 0).ravel()

            B = lambda xi, eta, zeta: torch.tensor([
                [-(eta - 1) * (zeta - 1), (eta - 1) * (zeta - 1), eta * (zeta - 1), -eta * (zeta - 1),
                    zeta * (eta - 1), -zeta * (eta - 1), -eta * zeta, eta * zeta],
                [-(xi - 1) * (zeta - 1), xi * (zeta - 1), (xi - 1) * (zeta - 1), -xi * (zeta - 1), zeta * (xi - 1),
                    -xi * zeta, -zeta * (xi - 1), xi * zeta],
                [-(eta - 1) * (xi - 1), xi * (eta - 1), eta * (xi - 1), -eta * xi, (eta - 1) * (xi - 1),
                    -xi * (eta - 1), -eta * (xi - 1), eta * xi]
            ], dtype=self.dtype)

            for i in range(n_gauss):
                grads = B(qx[i], qy[i], qz[i]) / h[:, None]
                grad_operators[i] = grads

        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")

        return grad_operators

    def get_grad_kernels(self) -> torch.Tensor:
        """
        Get kernels for symmetric gradient computation

        :return:
        """
        if self.grad_operator is None:
            self.grad_operator = self.get_grad_operator()

        if self.n_dim == 2:
            grad_kernels = self.grad_operator.reshape(-1, 2, 1, 2, 2).permute((0, 1, 2, 4, 3))
        elif self.n_dim == 3:
            grad_kernels = self.grad_operator.reshape(-1, 3, 1, 2, 2, 2).permute((0, 1, 2, 5, 4, 3))
        else:
            raise NotImplementedError(f"Dimension {self.n_dim} is not supported")

        return grad_kernels

    def get_div_kernels(self) -> torch.Tensor:
        """
        Get kernels for divergence computation

        :return:
        """
        if self.grad_kernels is None:
            self.grad_kernels = self.get_grad_kernels()
        grad_kernels = self.grad_kernels

        if self.n_dim == 2:
            div_kernels = torch.roll(grad_kernels.transpose(-3, -4), shifts=(1, 1), dims=(-2, -1))
        elif self.n_dim == 3:
            div_kernels = torch.roll(grad_kernels.transpose(-4, -5), shifts=(1, 1, 1), dims=(-3, -2, -1))

        return div_kernels
    
    def get_temp_loss(self, reduction: str="sum"):
        """
        Temperature field loss
        """
        return TempLoss(n_dim=self.n_dim, reduction=reduction)
    
    def get_flux_loss(self, reduction: str="sum"):
        """
        Flux field loss
        """         
        return FluxLoss(n_dim=self.n_dim, reduction=reduction)
    
    def compute_losses(self, field_pred: torch.Tensor, field_ref: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        Compute all relevant losses
        :param field_pred:
        :param field_ref:
        :param detach: detach grad
        :param cpu: transfer to cpu
        :return:
        """
        args = {"dtype": field_ref.dtype, "device": field_ref.device}
        field_pred = field_pred.to(**args)
        losses = {
            "temp": self.get_temp_loss(reduction="none")(field_pred, field_ref),
            "flux": self.get_flux_loss(reduction="none")(field_pred, field_ref),
        }
        return losses



import torch


class MatvecModule(torch.nn.Module):
    def __init__(self, n_dim, grad_module, model, div_module):
        super().__init__()
        self.n_dim = n_dim
        self.ch_dim = -(1 + self.n_dim)
        self.module_list = torch.nn.ModuleList([
            grad_module,
            model,
            div_module
        ])

    def forward(self, x, param_fields):
        batch_dims = x.shape[:self.ch_dim]
        x = torch.flatten(x, start_dim=0, end_dim=self.ch_dim - 1)
        x = self.module_list[0](x)  # grad
        x = torch.unflatten(x, dim=0, sizes=batch_dims)
        x = self.module_list[1](x, param_fields)[0]  # material law
        x = torch.flatten(x, start_dim=0, end_dim=self.ch_dim - 2)
        x = self.module_list[2](x)  # div
        x = torch.unflatten(x, dim=0, sizes=batch_dims)
        return x


class ResidualModule(torch.nn.Module):
    def __init__(self, n_dim, grad_module, model, div_module):
        super().__init__()
        self.n_dim = n_dim
        self.ch_dim = -(1 + self.n_dim)
        self.load_slice = (slice(None), None, slice(None)) + self.n_dim * (None,)
        self.module_list = torch.nn.ModuleList([
            grad_module,
            model,
            div_module
        ])

    def forward(self, x, param_fields, loadings):
        batch_dims = x.shape[:self.ch_dim]
        x = torch.flatten(x, start_dim=0, end_dim=self.ch_dim - 1)
        x = self.module_list[0](x)
        x = torch.unflatten(x, dim=0, sizes=batch_dims)
        x = x + loadings[self.load_slice].unsqueeze(0).expand(x.shape)
        x = self.module_list[1](x, param_fields)[0]
        x = torch.flatten(x, start_dim=0, end_dim=self.ch_dim - 2)
        x = -self.module_list[2](x)
        x = torch.unflatten(x, dim=0, sizes=batch_dims)
        return x


class RhsModule(torch.nn.Module):
    def __init__(self, n_dim, shape, model, div_module):
        super().__init__()
        self.n_dim = n_dim
        self.load_slice = (slice(None), None, slice(None)) + self.n_dim * (None,)
        self.shape = shape
        self.module_list = torch.nn.ModuleList([model, div_module])

    def forward(self, u, param_fields, loadings):
        batch_dims = (*param_fields.shape[:-(1 + self.n_dim)], loadings.shape[0])
        x = loadings[self.load_slice].expand(batch_dims + self.shape)
        x = self.module_list[0](x, param_fields)[0]
        x = torch.flatten(x, start_dim=0, end_dim=-(self.n_dim + 3))
        x = -self.module_list[1](x)
        x = torch.unflatten(x, dim=0, sizes=batch_dims)
        return x


class FieldModule(torch.nn.Module):
    def __init__(self, ch_dim, expand_dims, grad_module, model, padding_layers=[]):
        super().__init__()
        self.ch_dim = ch_dim
        self.expand_dims = expand_dims
        self.padding_layers = padding_layers
        self.module_list = torch.nn.ModuleList([grad_module, model])

    def forward(self, x, param_fields, loadings):
        batch_dims = x.shape[:self.ch_dim]
        y = torch.flatten(x, start_dim=0, end_dim=self.ch_dim - 1)
        y = self.module_list[0](y)  # gradient computation
        #print("after gradient", y.shape)
        y = torch.unflatten(y, dim=0, sizes=batch_dims)
        y = y + loadings[(slice(None), None, slice(None)) + self.expand_dims]  # construct strains
        y = self.module_list[1](y, param_fields)[0]  # material law
        y = y.mean(dim=self.ch_dim - 1)  # average over gauss points
        for padding_layer in self.padding_layers:
            x = padding_layer(x)
        return torch.cat([x, y], dim=self.ch_dim)



"""
Implementation of various iterative solvers in PyTorch

- CgSolver: Linear conjugate gradient solver
"""
from typing import Optional, Callable, Dict, Tuple

import torch

from unocg.problems import Problem
from unocg.solvers import Solver
from unocg.preconditioners import Preconditioner
from unocg.modules.solvers import CgFieldModule, CgModule


class CgSolver(Solver):
    """
    Linear conjugate gradient (CG) Solver
    """

    def __init__(
            self,
            problem: Problem,
            preconditioner: Optional[Preconditioner] = None,
            rtol: float = 1e-10,
            atol: float = 0.0,
            max_iter: int = 100000,
            verbose: int = 0,
            cg_type: str = "cg",
            log_residual: bool = False,
            log_iterates: bool = False,
            loss_callback: Optional[Callable[[torch.Tensor], Dict[str, torch.Tensor]]] = None,
    ):
        """

        :param problem:
        :param preconditioner:
        :param rtol:
        :param atol:
        :param max_iter:
        :param verbose:
        :param cg_type:
        :param log_residual:
        :param log_iterates:
        :param loss_callback:
        """
        super().__init__(problem, preconditioner, rtol, atol, max_iter, verbose)
        self.cg_type = cg_type
        self.log_residual = log_residual
        self.log_iterates = log_iterates
        self.loss_callback = loss_callback

    def solve(
            self,
            param_fields: torch.Tensor,
            loadings: Optional[torch.Tensor] = None,
            state_fields: Optional[torch.Tensor] = None,
            guess: Optional[torch.Tensor] = None,
            rtol: Optional[float] = None,
            atol: Optional[float] = None,
            max_iter: Optional[int] = None,
            verbose: Optional[int] = None,
            cg_type: Optional[str] = None,
            log_residual: Optional[bool] = None,
            log_iterates: Optional[bool] = None,
            loss_callback: Optional[Callable[[torch.Tensor], Dict[str, torch.Tensor]]] = None,
            zero_mean: bool = False,
            err_eps: float = 1e-10,
    ) -> Dict[str, torch.Tensor]:
        """
        Solve the problem for given parameter fields, loadings and state/history fields.

        :param param_fields: parameter fields
        :param loadings: loadings
        :param state_fields: initial state/history variables fields
        :param guess: initial guess for DOFs
        :param rtol: relative tolerance for residual norm
        :param atol: absolute tolerance for residual norm
        :param max_iter: maximum number of iterations
        :param cg_type: "cg" or "cgpr". More options will be implemented soon.
        :param log_residual: if residuals should be stored
        :param log_iterates: if iterates should be stored
        :param loss_callback: callback function for loss computation in every iteration
        :param zero_mean: if zero-mean property of iterates should be enforced
        :return: result data in form of a dictionary with keys
            "sol": DOF vectors u,
            "err_history": residual norm history,
            "n_iter": number of iterations,
            "state_fields": state/history variables after convergence
        """
        args = {"device": param_fields.device, "dtype": param_fields.dtype}

        if rtol is None:
            rtol = self.rtol
        if atol is None:
            atol = self.atol
        if max_iter is None:
            max_iter = self.max_iter
        if verbose is None:
            verbose = self.verbose
        if cg_type is None:
            cg_type = self.cg_type
        if log_residual is None:
            log_residual = self.log_residual
        if log_iterates is None:
            log_iterates = self.log_iterates
        if loss_callback is None:
            loss_callback = self.loss_callback

        if loadings is None:
            loadings = self.problem.loadings.to(**args)
        else:
            loadings = loadings.to(**args)
        n_loadings = loadings.shape[0]

        vec_shape = self.problem.get_vec_shape(param_fields, loadings=loadings)

        # Initial guess for temperature
        if guess is None:
            u = torch.zeros(vec_shape, **args)
        else:
            u = guess.reshape(vec_shape).to(**args)

        # Compute initial residual
        r, _ = self.problem.compute_residual(u, param_fields, loadings=loadings, state_fields=state_fields)

        # Norm of initial residual for convergence control
        err = torch.norm(r, torch.inf, dim=-1)
        err0 = err.clone()
        err_rel = torch.ones_like(err0)
        err_history = torch.unsqueeze(err0, dim=0)

        # Log metrics
        if log_residual:
            res_history = torch.unsqueeze(r.clone(), dim=0)
        if log_iterates:
            u_history = torch.unsqueeze(u.clone(), dim=0)
        if loss_callback is not None:
            losses = loss_callback(u)

        n_iter = 0
        d = torch.zeros_like(u)
        s = torch.zeros_like(u)
        delta = torch.ones_like(err0)
        if cg_type == "cgpr":
            s_old = torch.empty_like(s)

        while (err_rel.max() > rtol) and (err.max() > atol) and (n_iter < max_iter):
            # Determine "not converged" flags
            nc_all = torch.logical_and(err_rel > rtol, err > atol)
            if not nc_all.max():
                break  # everything is converged
            nc_load = nc_all.reshape(-1, n_loadings).max(dim=0)[0]
            nc = (Ellipsis, nc_load)
            ncd = nc + (slice(None),)

            if cg_type == "cgpr":
                s_old.copy_(s)

            # Apply preconditioner
            s[ncd] = self.preconditioner.apply(r[ncd]).reshape(r[ncd].shape)

            delta0 = delta.clone()
            delta = (r * s).sum(dim=-1)
            if cg_type == "cgpr":
                beta = torch.maximum((delta[nc] - (r[ncd] * s_old[ncd]).sum(dim=-1)) / delta0[nc], torch.zeros_like(err0[nc]))
            else:
                beta = delta[nc] / delta0[nc]

            # Update search direction
            d[ncd] = s[ncd] + beta[..., None] * d[ncd]

            # Compute matrix-vector product: p = A @ d
            p, _ = self.problem.matvec(d[ncd], param_fields=param_fields, state_fields=state_fields)
            alpha = delta[nc] / (d[ncd] * p).sum(dim=-1)

            # Update residual
            r[ncd] = r[ncd] - alpha[..., None] * p
            #r[ncd].sub_(alpha[..., None] * p)

            # Update solution approximation
            u[ncd] = u[ncd] + alpha[..., None] * d[ncd]
            #u[ncd].add_(alpha[..., None] * d[ncd])

            if zero_mean:
                u = self.problem.zero_mean(u)

            # Logging
            err = torch.norm(r, torch.inf, dim=-1)
            err_history = torch.cat([err_history, torch.unsqueeze(err, dim=0)], dim=0)
            if log_residual:
                res_history = torch.cat([res_history, torch.unsqueeze(r, dim=0)], dim=0)
            if log_iterates:
                u_history = torch.cat([u_history, torch.unsqueeze(u, dim=0)], dim=0)
            if loss_callback is not None:
                losses_i = loss_callback(u)
                losses = {loss_key: torch.cat([losses[loss_key], losses_]) for (loss_key, losses_) in losses_i.items()}
            err_rel = err / (err0 + err_eps)
            n_iter += 1

            if verbose > 0:
                print(f"iter {n_iter}, err {err.max().item()}, err_rel {err_rel.max().item()}")

        r, state_fields_new = self.problem.compute_residual(u, param_fields=param_fields, loadings=loadings, state_fields=state_fields)

        result = {"sol": u, "err_history": err_history, "n_iter": n_iter, "state_fields": state_fields_new}
        if log_residual:
            result["res_history"] = res_history
        if log_iterates:
            result["u_history"] = u_history
        if loss_callback is not None:
            result["losses"] = losses
        return result
    
    def get_module(self, dtype, device, **kwargs):
        args = {"dtype": dtype, "device": device}
        model = CgModule(
            n_channels=self.problem.n_channels,
            dof_shape=self.problem.dof_shape,
            rhs_model=self.problem.get_rhs_module(**args),
            matvec_model=self.problem.get_matvec_module(**args),
            prec_model=self.preconditioner.get_apply_module(**args),
            **kwargs
        )
        for param in model.parameters():
            param.requires_grad = False
        return model
    
    def get_field_module(self, dtype, device, **kwargs):
        args = {"dtype": dtype, "device": device}
        model = CgFieldModule(
            n_channels=self.problem.n_channels,
            dof_shape=self.problem.dof_shape,
            rhs_model=self.problem.get_rhs_module(**args),
            matvec_model=self.problem.get_matvec_module(**args),
            field_module=self.problem.get_field_module(**args),
            prec_model=self.preconditioner.get_apply_module(**args),
            **kwargs
        )
        for param in model.parameters():
            param.requires_grad = False
        return model



"""
Fourier-based Transforms
"""
import torch
import torch.nn.functional as F
from unocg.transforms import Transform
from unocg import Backend
from collections.abc import Iterable
import numpy as np
import math

try:
    import scipy as sp
except:
    pass

try:
    import pyfftw
    pyfftw.interfaces.cache.enable()
except:
    pass


class DiscreteFourierTransform(Transform):
    """
    Discrete Fourier Transform
    """
    def __init__(self, dim=-1, norm="backward", real=True, backend=Backend.TORCH):
        """
        Constructor

        :param n_dim:
        :param norm:
        :param real:
        :param backend:
        """
        super().__init__()
        if isinstance(dim, Iterable):
            self.dim = dim
        else:
            self.dim = [dim]
        self.norm = norm
        self.real = real
        self.backend = backend

        if self.backend == Backend.TORCH:
            if self.real:
                self.fftn = torch.fft.rfftn
                self.ifftn = torch.fft.irfftn
            else:
                self.fftn = torch.fft.fftn
                self.ifftn = torch.fft.ifftn
        elif self.backend == Backend.SCIPY:
            try:
                if self.real:
                    self.fftn = sp.fft.rfftn
                    self.ifftn = sp.fft.irfftn
                else:
                    self.fftn = sp.fft.fftn
                    self.ifftn = sp.fft.ifftn
            except ImportError:
                raise ImportError("SciPy installation not found, but was specified as backend")
        elif self.backend == Backend.FFTW:
            try:
                self.pyfftw_threads = 16
                if self.real:

                    def fftn(x, dim, *args, **kwargs):
                        if isinstance(x, torch.Tensor):
                            x = x.detach().cpu().numpy()
                        return pyfftw.interfaces.numpy_fft.rfftn(x, axes=dim, *args, **kwargs, threads=self.pyfftw_threads)
                    
                    self.fftn = fftn

                    def ifftn(x, dim, *args, **kwargs):
                        if isinstance(x, torch.Tensor):
                            x = x.detach().cpu().numpy()
                        return pyfftw.interfaces.numpy_fft.irfftn(x, axes=dim, *args, **kwargs, threads=self.pyfftw_threads)
                    
                    self.ifftn = ifftn
                else:

                    def fftn(x, dim, *args, **kwargs):
                        if isinstance(x, torch.Tensor):
                            x = x.detach().cpu().numpy()
                        return pyfftw.interfaces.numpy_fft.fftn(x, axes=dim, *args, **kwargs, threads=self.pyfftw_threads)
                    
                    self.fftn = fftn

                    def ifftn(x, dim, *args, **kwargs):
                        if isinstance(x, torch.Tensor):
                            x = x.detach().cpu().numpy()
                        return pyfftw.interfaces.numpy_fft.ifftn(x, axes=dim, *args, **kwargs, threads=self.pyfftw_threads)
                    
                    self.ifftn = ifftn
            except ImportError:
                raise ImportError("PyFFTW installation not found, but was specified as backend")
        else:
            raise NotImplementedError(f"Backend {self.backend} is not implemented for this operation")

    def transform(self, x, s=None):
        s_dim = None if s is None else tuple(np.atleast_1d(np.atleast_1d(s)[self.dim]))
        return self.fftn(x, dim=self.dim, s=s_dim)

    def inverse_transform(self, x, s=None):
        s_dim = None if s is None else tuple(np.atleast_1d(np.atleast_1d(s)[self.dim]))
        return self.ifftn(x, dim=self.dim, s=s_dim)
    
    def adjoint_transform(self, x, s=None):
        s_dim = None if s is None else np.atleast_1d(np.atleast_1d(s)[self.dim])
        return self.inverse_transform(x, s=s_dim)


class DiscreteSineTransform(Transform):
    """
    Discrete Sine Transform
    """
    def __init__(self, dim=-1, norm="ortho", type=1, backend=Backend.TORCH):
        """
        Constructor

        :param n_dim:
        :param norm:
        :param type:
        :param backend:
        """
        super().__init__()
        if isinstance(dim, Iterable):
            self.dim = dim
        else:
            self.dim = [dim]
        self.type = type
        self.backend = backend
        self.norm = norm
        self.rfft_transform = DiscreteFourierTransform(real=True, norm=self.norm, dim=-1, backend=self.backend)
        self.pyfftw_threads = 16

        if self.type != 1:
            raise NotImplementedError(f"Discrete Sine Transform of type {self.type} not implemented")

    def transform(self, x, s=None):
        s = None if s is None else np.atleast_1d(s)
        for dim in self.dim:
            s_dim = None if s is None else np.atleast_1d(s[dim])
            x = self.dst_1d(x, dim=dim, s=s_dim)
        return x

    def inverse_transform(self, x, s=None):
        s = None if s is None else np.atleast_1d(s)
        for dim in self.dim:
            s_dim = None if s is None else np.atleast_1d(s[dim])
            x = self.idst_1d(x, dim=dim, s=s_dim)
        return x
    
    def adjoint_transform(self, x, s=None):
        s = None if s is None else np.atleast_1d(s)
        return self.transform(x)

    def dst_1d(self, x, s=None, dim=-1):
        if self.backend == Backend.TORCH:
            n = x.shape[dim]
            padded_x = torch.tensor(-1j, device=x.device) * F.pad(x.swapaxes(dim, -1), (1, 1))
            padded_x_hat = self.rfft_transform.inverse_transform(padded_x)
            x_hat = padded_x_hat[...,1:n+1].real.swapaxes(-1, dim) * math.sqrt(2 * (n + 1))
        elif self.backend == Backend.FFTW:
            x_hat = pyfftw.interfaces.scipy_fft.dstn(x, axes=dim, type=self.type, norm=self.norm, workers=self.pyfftw_threads)
        elif self.backend == Backend.SCIPY:
            x_hat = sp.fft.dstn(x, axes=dim, type=self.type, norm=self.norm)
        else:
            raise NotImplementedError(f"Backend {self.backend} is not supported for this operation")
        return x_hat

    def idst_1d(self, x, s=None, dim=-1):
        if self.backend == Backend.TORCH:
            n = x.shape[dim]
            x_hat = self.dst_1d(x, s=s, dim=dim)
        elif self.backend == Backend.FFTW:
            x_hat = pyfftw.interfaces.scipy_fft.idstn(x, axes=dim, type=self.type, norm=self.norm, workers=self.pyfftw_threads)
        elif self.backend == Backend.SCIPY:
            x_hat = sp.fft.idstn(x, axes=dim, type=self.type, norm=self.norm, workers=self.pyfftw_threads)
        else:
            raise NotImplementedError(f"Backend {self.backend} is not supported for this operation")
        return x_hat


class DiscreteCosineTransform(Transform):
    """
    Discrete Cosine Transform
    """
    def __init__(self, dim=-1, norm="ortho", type=1, backend=Backend.TORCH):
        """
        Constructor

        :param n_dim:
        :param norm:
        :param type:
        """
        super().__init__()
        if isinstance(dim, Iterable):
            self.dim = dim
        else:
            self.dim = [dim]
        self.type = type
        self.norm = norm
        self.backend = backend
        self.pyfftw_threads = 16

        if self.backend == Backend.TORCH:
            self.rfft_transform = DiscreteFourierTransform(real=True, norm=self.norm, dim=-1, backend=self.backend)

        if self.type != 1:
            raise NotImplementedError(f"Discrete Cosine Transform of type {self.type} not implemented")

    def transform(self, x, s=None):
        for dim in self.dim:
            x = self.dct_1d(x, dim=dim, s=s)
        return x

    def inverse_transform(self, x, s=None):
        for dim in self.dim:
            x = self.idct_1d(x, dim=dim, s=s)
        return x

    def adjoint_transform(self, x, s=None):
        # TODO: check if this is correct!
        return self.transform(x, s=s)

    def dct_1d(self, x, s=None, dim=-1):
        """
        Discrete Cosine Transform, Type I in 1D

        :param x: the input signal
        :return: the DCT-I of the signal over the last dimension
        """
        if self.backend == Backend.TORCH:
            if self.norm != "ortho":
                raise NotImplementedError(f"Norm {norm} not implemented")
            xt = x.swapaxes(dim, -1)
            xt[..., 0] *= math.sqrt(2)
            xt[..., -1] *= math.sqrt(2)
            padded_xt = torch.cat([xt, xt.flip([-1])[..., 1:-1]], dim=-1)
            xt_hat = self.rfft_transform.transform(padded_xt).real
            xt_hat[..., 0] /= math.sqrt(2)
            xt_hat[..., -1] /= math.sqrt(2)
            xt_hat /= math.sqrt(2 * (xt.shape[-1] - 1))
            x_hat = xt_hat.swapaxes(dim, -1)
        elif self.backend == Backend.FFTW:
            x_hat = pyfftw.interfaces.scipy_fft.dctn(x, axes=dim, type=self.type, norm=self.norm, workers=self.pyfftw_threads)
        elif self.backend == Backend.SCIPY:
            x_hat = torch.tensor(sp.fft.dctn(x.cpu().numpy(), axes=dim, type=self.type, norm=self.norm))
        else:
            raise NotImplementedError(f"Backend {self.backend} is not supported for this operation")
        return x_hat

    def idct_1d(self, x, s=None, dim=-1):
        """
        The inverse of DCT-I in 1D, which is just a scaled DCT-I

        :param X: the input signal
        :return: the inverse DCT-I of the signal over the last dimension
        """
        if self.backend == Backend.TORCH:
            if self.norm != "ortho":
                raise NotImplementedError(f"Norm {norm} not implemented")
            x_hat = self.dct_1d(x, s=s, dim=dim)
        elif self.backend == Backend.FFTW:
            x_hat = pyfftw.interfaces.scipy_fft.idctn(x, axes=dim, type=self.type, norm=self.norm, workers=self.pyfftw_threads)
        elif self.backend == Backend.SCIPY:
            x_hat = torch.tensor(sp.fft.idctn(x.cpu().numpy(), axes=dim, type=self.type, norm=self.norm))
        else:
            raise NotImplementedError(f"Backend {self.backend} is not supported for this operation")
        return x_hat


class DiscreteMixedTransform(Transform):
    """
    Discrete Mixed Transform
    """
    def __init__(self, dim_fourier=None, dim_sine=None, dim_cosine=None, real=True, norm="backward", type=1, backend=Backend.TORCH):
        """
        Constructor

        :param n_dim:
        :param norm:
        :param type:
        """
        super().__init__()
        self.type = type
        self.norm = norm

        if dim_fourier is None:
            self.dim_fourier = []
        else:
            self.dim_fourier = dim_fourier

        if dim_sine is None:
            self.dim_sine = []
        else:
            self.dim_sine = dim_sine

        if dim_cosine is None:
            self.dim_cosine = []
        else:
            self.dim_cosine = dim_cosine
        
        self.rfft_transform = DiscreteFourierTransform(real=real, norm=norm, dim=self.dim_fourier, backend=backend)
        self.dst_transform = DiscreteSineTransform(norm=norm, dim=self.dim_sine, backend=backend)
        self.dct_transform = DiscreteCosineTransform(norm=norm, dim=self.dim_cosine, backend=backend)

        if self.type != 1:
            raise NotImplementedError(f"Discrete Mixed Transform of type {self.type} not implemented")

    def transform(self, x, s=None):
        s = None if s is None else np.atleast_1d(s)
        x = self.dst_transform.transform(x, s=s)
        x = self.dct_transform.transform(x, s=s)
        x = self.rfft_transform.transform(x, s=s)
        return x

    def inverse_transform(self, x, s=None):
        s = None if s is None else np.atleast_1d(s)
        x = self.rfft_transform.inverse_transform(x, s=s)
        x = self.dct_transform.inverse_transform(x, s=s)
        x = self.dst_transform.inverse_transform(x, s=s)
        return x

    def adjoint_transform(self, x, s=None):
        s = None if s is None else np.atleast_1d(s)
        x = self.rfft_transform.adjoint_transform(x, s=s)
        x = self.dct_transform.adjoint_transform(x, s=s)
        x = self.dst_transform.adjoint_transform(x, s=s)
        return x



"""
Preconditioner base class.
"""
import torch
import numpy as np

from unocg.problems import Problem


class Preconditioner:
    """
    Generic Preconditioner
    """
    def __init__(self, problem: Problem):
        self.problem = problem

    def apply(self, r: torch.Tensor) -> torch.Tensor:
        """
        Apply preconditioner

        :param r:
        :return:
        """
        return r.clone()

    def apply_field(self, res_field: torch.Tensor) -> torch.Tensor:
        r = self.problem.reshape_vec(res_field)
        s = self.apply(r)
        return self.problem.reshape_field(s)

    def __call__(self, r: torch.Tensor) -> torch.Tensor:
        return self.apply(r)

    def get_apply_module(self, dtype, device):
        """
        Build PyTorch model for apply operation

        Input shape for model: [B, L, C, N, N] or [B, L, C, N, N, N]
        Output shape: same as input shape
        """
        return torch.nn.Identity()
